from __future__ import annotations

from typing import Dict, List, Union

import numpy as np
import pandas as pd
from gbd_mapping import (
    Cause,
    Covariate,
    Etiology,
    ModelableEntity,
    RiskFactor,
    Sequela,
    causes,
)

from vivarium_inputs import utilities, utility_data
from vivarium_inputs.globals import (
    BOUNDARY_SPECIAL_CASES,
    DRAW_COLUMNS,
    PROTECTIVE_CAUSE_RISK_PAIRS,
    RISKS_WITH_NEGATIVE_PAF,
    DataTransformationError,
    Population,
    gbd,
)
from vivarium_inputs.mapping_extension import (
    AlternativeRiskFactor,
    HealthcareEntity,
    HealthTechnology,
)
from vivarium_inputs.utilities import DataType
from vivarium_inputs.validation.shared import check_value_columns_boundary

VALID_INCIDENCE_RANGE = (0.0, 50.0)
VALID_PREVALENCE_RANGE = (0.0, 1.0)
VALID_BIRTH_PREVALENCE_RANGE = (0.0, 1.0)
VALID_DISABILITY_WEIGHT_RANGE = (0.0, 1.0)
VALID_REMISSION_RANGE = (0.0, 120.0)  # James' head
# FIXME: bumping csmr max because of age group scaling 2/8/19 - K.W.
VALID_CAUSE_SPECIFIC_MORTALITY_RANGE = (
    0.0,
    6,
)  # used mortality viz, picked worst country 15q45, mul by ~1.25
VALID_EXCESS_MORT_RANGE = (
    0.0,
    300.0,
)  # subarachnoid_hemorrhage has em values close to 300 for neonatal age groups
VALID_EXPOSURE_RANGE = (0.0, {"continuous": 10_000.0, "categorical": 1.0})
VALID_EXPOSURE_SD_RANGE = (0.0, 1000.0)  # James' brain
VALID_EXPOSURE_DIST_WEIGHTS_RANGE = (0.0, 1.0)
# upper limit of 1000 chosen by Abie after seeing RRs of 950 for sbp
VALID_RELATIVE_RISK_RANGE = (1.0, {"continuous": 1000.0, "categorical": 2550.0})
VALID_PAF_RANGE = (0.0, 1.0)
VALID_PROTECTIVE_PAF_MIN = -1.0
VALID_COST_RANGE = (0, {"healthcare_entity": 30_000, "health_technology": 50})
# FIXME: bumping for Mexico (max utilization ~422) 9/12/19 - K.W.
VALID_UTILIZATION_RANGE = (0, 500)
VALID_POPULATION_RANGE = (
    0,
    360_000_000_000,
)  # Upper bound pulled from GBD global population largest sex-age bin.
VALID_LIFE_EXP_RANGE = (0, 90)

SCRUBBED_DEMOGRAPHIC_COLUMNS = ["location", "sex", "age", "year"]


class SimulationValidationContext:
    def __init__(self, location: list[str], **additional_data):
        self.context_data = {"location": location}
        self.context_data.update(additional_data)

        if "years" not in self.context_data:
            self.context_data["years"] = utility_data.get_year_block()
        if "age_bins" not in self.context_data:
            self.context_data["age_bins"] = utility_data.get_age_bins()

    def __getitem__(self, key):
        return self.context_data[key]

    def __setitem__(self, key, value):
        self.context_data[key] = value


def validate_for_simulation(
    data: pd.DataFrame,
    entity: ModelableEntity,
    measure: str,
    location: int | str | list[int | str],
    years: int | None,
    value_columns: list[str],
    **context_args,
) -> None:
    """Validate data for use in a simulation.

    Checks that the data conforms to the format that is expected by the simulation
    as well as conforms to normal expectations for a measure.

    Data coming in to the simulation is expected to have a full demographic set
    in most instances, as well non-missing, non-infinite, reasonable data. This
    function enforces column names, the demographic extents, and expected
    ranges and relationships of measure data.

    The following standard checks are applied:
    1. Validate standard columns and indices:
        For all demographic columns, ensure that the column names are
        correct and the values in the columns match the expected set contained
        in the given context.
    2. Validate value columns:
        Ensure that the column name is correct and check that all values within
        the column fall in the expected range.
    3. Validate age/sex restrictions if applicable:
        Ensure that the data matches any age or sex restrictions of the entity.

    Parameters
    ----------
    data
        Data to be validated.
    entity
        The GBD Entity to which the data pertain.
    measure
        The measure to which the data pertain.
    location
        The location(s) to which the data pertain.
    check_all_years
        Flag indicating whether to validate that we have all years.
        Otherwise, validate that data has most recent year.
        Defaults to False.
    value_columns
        List of column names in `data` that contain the values to be validated.
    **context_args
        Any data or information needed to construct the SimulationContext used
        by the individual entity-measure validator functions.

    Raises
    -------
    DataTransformationError
        If any columns are mis-formatted or assumptions about the data are
        violated.
    """
    validators = {
        # Cause-like measures
        "incidence_rate": validate_incidence_rate,
        "prevalence": validate_prevalence,
        "birth_prevalence": validate_birth_prevalence,
        "disability_weight": validate_disability_weight,
        "remission_rate": validate_remission_rate,
        "cause_specific_mortality_rate": validate_cause_specific_mortality_rate,
        "excess_mortality_rate": validate_excess_mortality_rate,
        # Risk-like measures
        "exposure": validate_exposure,
        "exposure_standard_deviation": validate_exposure_standard_deviation,
        "exposure_distribution_weights": validate_exposure_distribution_weights,
        "relative_risk": validate_relative_risk,
        "population_attributable_fraction": validate_population_attributable_fraction,
        "mediation_factors": validate_mediation_factors,
        # Covariate measures
        "estimate": validate_estimate,
        # Population measures
        "structure": validate_structure,
        "theoretical_minimum_risk_life_expectancy": validate_theoretical_minimum_risk_life_expectancy,
        "age_bins": validate_age_bins,
        "demographic_dimensions": validate_demographic_dimensions,
    }

    if measure not in validators:
        raise NotImplementedError(f"Validation for measure '{measure}' not implemented.")

    if years != "all":
        if isinstance(years, int):
            context_args["years"] = pd.DataFrame(
                {"year_start": years, "year_end": years + 1}, index=[0]
            )
        elif isinstance(years, list):
            context_args["years"] = pd.DataFrame(
                {
                    "year_start": sorted(years),
                    "year_end": [year + 1 for year in sorted(years)],
                }
            )
        else:
            most_recent_year = utility_data.get_most_recent_year()
            context_args["years"] = pd.DataFrame(
                {"year_start": most_recent_year, "year_end": most_recent_year + 1}, index=[0]
            )

    # Coerce to location names
    if not isinstance(location, list):
        location = [location]
    location = [
        utility_data.get_location_name(loc) if isinstance(loc, int) else loc
        for loc in location
    ]
    context = SimulationValidationContext(location, **context_args)
    validators[measure](data, entity, context, value_columns)


#########################################################
#   VALIDATE SIM DATA ENTITY-MEASURE SPECIFIC METHODS   #
# ----------------------------------------------------- #
#   Signatures to match as used in validate_sim_data    #
#########################################################


def validate_incidence_rate(
    data: pd.DataFrame,
    entity: Cause | Sequela,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped incidence
    data.

    Parameters
    ----------
    data
        Simulation-prepped incidence data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yld age or sex restrictions are violated, or data falls outside the
        expected boundary values.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, value_columns, data.columns
    )

    validate_standard_columns(data, context)

    check_value_columns_boundary(
        data,
        boundary_value=VALID_INCIDENCE_RANGE[0],
        boundary_type="lower",
        value_columns=value_columns,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        boundary_value=VALID_INCIDENCE_RANGE[1],
        boundary_type="upper",
        value_columns=value_columns,
        error=DataTransformationError,
    )

    restrictions_entity = (
        [c for c in causes if entity in c.sequelae][0] if entity.kind == "sequela" else entity
    )
    check_age_restrictions(
        data, restrictions_entity, rest_type="yld", fill_value=0.0, context=context
    )
    check_sex_restrictions(
        data,
        restrictions_entity.restrictions.male_only,
        restrictions_entity.restrictions.female_only,
        fill_value=0.0,
    )


def validate_prevalence(
    data: pd.DataFrame,
    entity: Cause | Sequela,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped prevalence data.

    Parameters
    ----------
    data
        Simulation-prepped prevalence data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yld age or sex restrictions are violated, or data falls outside the
        expected boundary values.
    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, value_columns, data.columns
    )

    validate_standard_columns(data, context)

    check_value_columns_boundary(
        data,
        boundary_value=VALID_PREVALENCE_RANGE[0],
        boundary_type="lower",
        value_columns=value_columns,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        boundary_value=VALID_PREVALENCE_RANGE[1],
        boundary_type="upper",
        value_columns=value_columns,
        error=DataTransformationError,
    )

    restrictions_entity = (
        [c for c in causes if entity in c.sequelae][0] if entity.kind == "sequela" else entity
    )
    check_age_restrictions(
        data, restrictions_entity, rest_type="yld", fill_value=0.0, context=context
    )
    check_sex_restrictions(
        data,
        restrictions_entity.restrictions.male_only,
        restrictions_entity.restrictions.female_only,
        fill_value=0.0,
    )


def validate_birth_prevalence(
    data: pd.DataFrame,
    entity: Cause | Sequela,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped birth
    prevalence data, skipping the check on age columns since birth prevalence
    is not age-specific.

    Parameters
    ----------
    data
        Simulation-prepped birth prevalence data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if sex restrictions are violated, or data falls outside the expected
        boundary values.

    """
    expected_index_names = ["location", "sex", "year_start", "year_end"]
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    validate_location_column(data, context)
    validate_sex_column(data)
    validate_year_column(data, context)
    validate_value_column(data)

    check_value_columns_boundary(
        data,
        boundary_value=VALID_BIRTH_PREVALENCE_RANGE[0],
        boundary_type="lower",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        boundary_value=VALID_BIRTH_PREVALENCE_RANGE[1],
        boundary_type="upper",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )

    restrictions_entity = (
        [c for c in causes if entity in c.sequelae][0] if entity.kind == "sequela" else entity
    )
    check_sex_restrictions(
        data,
        restrictions_entity.restrictions.male_only,
        restrictions_entity.restrictions.female_only,
        fill_value=0.0,
    )


def validate_disability_weight(
    data: pd.DataFrame,
    entity: Cause | Sequela,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped disability
    weight data.

    Parameters
    ----------
    data
        Simulation-prepped disability weight data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yld age or sex restrictions are violated, or data falls outside the
        expected boundary values.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    validate_standard_columns(data, context)
    check_value_columns_boundary(
        data,
        boundary_value=VALID_DISABILITY_WEIGHT_RANGE[0],
        boundary_type="lower",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        boundary_value=VALID_DISABILITY_WEIGHT_RANGE[1],
        boundary_type="upper",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )

    restrictions_entity = (
        [c for c in causes if entity in c.sequelae][0] if entity.kind == "sequela" else entity
    )
    check_age_restrictions(
        data, restrictions_entity, rest_type="yld", fill_value=0.0, context=context
    )
    check_sex_restrictions(
        data,
        restrictions_entity.restrictions.male_only,
        restrictions_entity.restrictions.female_only,
        fill_value=0.0,
    )


def validate_remission_rate(
    data: pd.DataFrame,
    entity: Cause,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped remission
    data.

    Parameters
    ----------
    data
        Simulation-prepped remission data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yld age or sex restrictions are violated, or data falls outside the
        expected boundary values.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    validate_standard_columns(data, context)

    check_value_columns_boundary(
        data,
        boundary_value=VALID_REMISSION_RANGE[0],
        boundary_type="lower",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        boundary_value=VALID_REMISSION_RANGE[1],
        boundary_type="upper",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )

    check_age_restrictions(data, entity, rest_type="yld", fill_value=0.0, context=context)
    check_sex_restrictions(
        data, entity.restrictions.male_only, entity.restrictions.female_only, fill_value=0.0
    )


def validate_cause_specific_mortality_rate(
    data: pd.DataFrame,
    entity: Cause,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped cause
    specific mortality data.

    Parameters
    ----------
    data
        Simulation-prepped cause specific mortality data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yll age or sex restrictions are violated, or data falls outside the
        expected boundary values.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    validate_standard_columns(data, context)

    check_value_columns_boundary(
        data,
        boundary_value=VALID_CAUSE_SPECIFIC_MORTALITY_RANGE[0],
        boundary_type="lower",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        boundary_value=VALID_CAUSE_SPECIFIC_MORTALITY_RANGE[1],
        boundary_type="upper",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )

    check_age_restrictions(data, entity, rest_type="yll", fill_value=0.0, context=context)
    check_sex_restrictions(
        data, entity.restrictions.male_only, entity.restrictions.female_only, fill_value=0.0
    )


def validate_excess_mortality_rate(
    data: pd.DataFrame,
    entity: Cause,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped excess
    mortality data.

    Parameters
    ----------
    data
        Simulation-prepped excess mortality data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yll age or sex restrictions are violated, or data falls outside the
        expected boundary values.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    validate_standard_columns(data, context)

    check_value_columns_boundary(
        data,
        boundary_value=VALID_EXCESS_MORT_RANGE[0],
        boundary_type="lower",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )

    for location in context["location"]:
        if entity.name in BOUNDARY_SPECIAL_CASES["excess_mortality_rate"].get(location, {}):
            max_val = BOUNDARY_SPECIAL_CASES["excess_mortality_rate"][context["location"]][
                entity.name
            ]
        else:
            max_val = VALID_EXCESS_MORT_RANGE[1]
        check_value_columns_boundary(
            data,
            boundary_value=max_val,
            boundary_type="upper",
            value_columns=DRAW_COLUMNS,
            error=DataTransformationError,
        )

    check_age_restrictions(data, entity, rest_type="yll", fill_value=0.0, context=context)
    check_sex_restrictions(
        data, entity.restrictions.male_only, entity.restrictions.female_only, fill_value=0.0
    )


def validate_exposure(
    data: pd.DataFrame,
    entity: RiskFactor | AlternativeRiskFactor,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped exposure
    data, with the upper boundary of values determined by the distribution type.
    The broadest age range determined by yll and yld age restrictions is used
    in the restriction check. For categorical risks, because exposure is filled
    with two values outside the restriction boundaries (one for the TMREL
    category and one for all other categories), check that only these two
    values are found outside restrictions and that they are matched to the
    right categories.

    Additionally, check that the parameter column contains either only
    'continuous' for an entity with a continuous distribution or the full set
    of categories for an entity with a categorical distribution. For entities
    with categorical distributions, further check that exposure values sum to
    one across categories for all demographic groups.

    Parameters
    ----------
    data
        Simulation-prepped exposure data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if the broadest age or sex restrictions are violated, data falls
        outside the expected boundary values, the parameter column does not
        contain the expected values based on the entity's distribution type, or
        exposure values do not sum to 1 for entities with a categorical
        distribution.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS + ["parameter"]
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    is_continuous = entity.distribution in ["normal", "lognormal", "ensemble"]
    is_categorical = entity.distribution in [
        "dichotomous",
        "ordered_polytomous",
        "unordered_polytomous",
    ]

    if is_continuous:
        if set(data.index.unique("parameter")) != {"continuous"}:
            raise DataTransformationError(
                "Continuous exposure data should contain "
                "'continuous' in the parameter column."
            )
        valid_kwd = "continuous"
    elif is_categorical:
        if set(data.index.unique("parameter")) != set(
            entity.categories.to_dict()
        ):  # to_dict removes None
            raise DataTransformationError(
                "Categorical exposure data does not contain all "
                "categories in the parameter column."
            )
        valid_kwd = "categorical"
    else:
        raise NotImplementedError()

    check_value_columns_boundary(
        data,
        boundary_value=VALID_EXPOSURE_RANGE[0],
        boundary_type="lower",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        boundary_value=VALID_EXPOSURE_RANGE[1][valid_kwd],
        boundary_type="upper",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )

    cats = data.groupby("parameter")
    cats.apply(validate_standard_columns, context)

    if is_categorical:
        non_categorical_columns = list(set(data.index.names).difference({"parameter"}))
        if not np.allclose(data.groupby(non_categorical_columns).sum(), 1.0):
            raise DataTransformationError(
                "Categorical exposures do not sum to one across categories."
            )

    if entity.kind in ["risk_factor", "alternative_risk_factor"]:
        fill_value = {"exposed": 0.0, "unexposed": 1.0} if is_categorical else 0.0
        check_age_restrictions(
            data, entity, rest_type="outer", fill_value=fill_value, context=context
        )
        check_sex_restrictions(
            data,
            entity.restrictions.male_only,
            entity.restrictions.female_only,
            fill_value=fill_value,
            entity=entity,
        )


def validate_exposure_standard_deviation(
    data: pd.DataFrame,
    entity: RiskFactor | AlternativeRiskFactor,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped exposure
    standard deviation data, using the broadest age range determined by yll and
    yld age restrictions in the restriction check.

    Parameters
    ----------
    data
        Simulation-prepped exposure standard deviation data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if the broadest age or sex restrictions are violated, or data falls
        outside the expected boundary values.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    validate_standard_columns(data, context)

    check_value_columns_boundary(
        data,
        boundary_value=VALID_EXPOSURE_SD_RANGE[0],
        boundary_type="lower",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        boundary_value=VALID_EXPOSURE_SD_RANGE[1],
        boundary_type="upper",
        value_columns=DRAW_COLUMNS,
        error=DataTransformationError,
    )

    check_age_restrictions(data, entity, rest_type="outer", fill_value=0.0, context=context)
    check_sex_restrictions(
        data, entity.restrictions.male_only, entity.restrictions.female_only, fill_value=0.0
    )


def validate_exposure_distribution_weights(
    data: pd.DataFrame,
    entity: RiskFactor | AlternativeRiskFactor,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped exposure
    distribution weights data, skipping the check on draw columns since weights
    data is not by draw and using the broadest age range determined by yll
    and yld age restrictions in the restriction check.

    Additionally, ensure that weights sum to 1 in all demographic groups
    expected to have valid weights (i.e., all age groups included in
    corresponding exposure data) and to 0 in all other demographic groups.

    Parameters
    ----------
    data
        Simulation-prepped exposure distribution weights data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if the broadest age or sex restrictions are violated, data falls
        outside the expected boundary values, or weights don't sum to 1 or 0.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS + ["parameter"]
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, ["value"], data.columns
    )

    validate_demographic_columns(data, context)
    validate_value_column(data)

    check_value_columns_boundary(
        data,
        boundary_value=VALID_EXPOSURE_DIST_WEIGHTS_RANGE[0],
        boundary_type="lower",
        value_columns=["value"],
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        boundary_value=VALID_EXPOSURE_DIST_WEIGHTS_RANGE[1],
        boundary_type="upper",
        value_columns=["value"],
        error=DataTransformationError,
    )

    non_weight_columns = SCRUBBED_DEMOGRAPHIC_COLUMNS
    weights_sum = data.groupby(non_weight_columns)["value"].sum()
    if not weights_sum.apply(lambda s: np.isclose(s, 1.0) or np.isclose(s, 0.0)).all():
        raise DataTransformationError(
            "Exposure weights do not sum to one across demographics."
        )

    check_age_restrictions(data, entity, rest_type="outer", fill_value=0.0, context=context)
    check_sex_restrictions(
        data, entity.restrictions.male_only, entity.restrictions.female_only, fill_value=0.0
    )


def validate_relative_risk(
    data: pd.DataFrame,
    entity: RiskFactor,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped relative risk
    data, with the upper boundary of values determined by the distribution type.
    Because some RRs may be protective, allow the lower boundary to be 0 for
    those special cases. Otherwise, check a lower boundary of 1. Use the
    affected_measure to determine which set of age restrictions to check:
    using yll for excess mortality and the narrowest range for incidence rate.

    Additionally, for entities with categorical distributions, check that the
    theoretical minimum risk exposure level (TMREL) category contains exposure
    values of only 1.

    Parameters
    ----------
    data
        Simulation-prepped relative risk data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if the age or sex restrictions are violated, data falls
        outside the expected boundary values, or TMREL values are not all 1 for
        entities with categorical distributions.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS + [
        "affected_entity",
        "affected_measure",
        "parameter",
    ]
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    risk_relationship = data.groupby(["affected_entity", "affected_measure", "parameter"])
    risk_relationship.apply(validate_standard_columns, context)

    is_continuous = entity.distribution in ["normal", "lognormal", "ensemble"]
    is_categorical = entity.distribution in [
        "dichotomous",
        "ordered_polytomous",
        "unordered_polytomous",
    ]

    if is_categorical:
        range_kwd = "categorical"
    elif is_continuous:
        range_kwd = "continuous"
    else:
        raise NotImplementedError()

    protective_causes = [c.name for c in PROTECTIVE_CAUSE_RISK_PAIRS.get(entity.name, [])]
    protective_mask = data.index.isin(protective_causes, "affected_entity")

    if np.any(protective_mask):
        check_value_columns_boundary(
            data.loc[protective_mask],
            boundary_value=0,
            boundary_type="lower",
            value_columns=DRAW_COLUMNS,
            inclusive=False,
            error=DataTransformationError,
        )
        check_value_columns_boundary(
            data.loc[protective_mask],
            boundary_value=VALID_RELATIVE_RISK_RANGE[0],
            boundary_type="upper",
            value_columns=DRAW_COLUMNS,
        )

    non_protective_mask = ~protective_mask
    if np.any(non_protective_mask):
        check_value_columns_boundary(
            data.loc[non_protective_mask],
            boundary_value=VALID_RELATIVE_RISK_RANGE[0],
            boundary_type="lower",
            value_columns=DRAW_COLUMNS,
        )

        check_value_columns_boundary(
            data.loc[non_protective_mask],
            boundary_value=VALID_RELATIVE_RISK_RANGE[1][range_kwd],
            boundary_type="upper",
            value_columns=DRAW_COLUMNS,
            error=DataTransformationError,
        )

    if is_categorical:
        tmrel_cat = utility_data.get_tmrel_category(entity)
        if not (data.loc[data.index.isin([tmrel_cat], "parameter")].values == 1.0).all():
            raise DataTransformationError(
                f"The TMREL category {tmrel_cat} contains values other than 1.0."
            )

    if entity.kind in ["risk_factor", "alternative_risk_factor"]:
        if (data.index.unique("affected_measure") == "incidence_rate").all():
            check_age_restrictions(
                data, entity, rest_type="inner", fill_value=1.0, context=context
            )
        else:
            check_age_restrictions(
                data, entity, rest_type="yll", fill_value=1.0, context=context
            )
        check_sex_restrictions(
            data,
            entity.restrictions.male_only,
            entity.restrictions.female_only,
            fill_value=1.0,
        )


def validate_population_attributable_fraction(
    data: pd.DataFrame,
    entity: RiskFactor | Etiology,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped population
    attributable fraction data. For protective cause-risk pairs: check a hard
    lower boundary of -1, a soft upper bound of 0 and a hard upper bound of 1.
    For non-protective cause-risk pairs: check a hard lower bound of 0 and a
    hard upper bound of 1. Check age and sex restrictions based on the affected
    entity and measure.

    Parameters
    ----------
    data
        Simulation-prepped population attributable fraction data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if the age or sex restrictions are violated, or data falls
        outside the expected boundary values.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS + [
        "affected_entity",
        "affected_measure",
    ]
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    risk_relationship = data.groupby(["affected_entity", "affected_measure"])
    risk_relationship.apply(validate_standard_columns, context)

    protective_causes = [c.name for c in PROTECTIVE_CAUSE_RISK_PAIRS.get(entity.name, [])]
    protective = data.loc[data.index.isin(protective_causes, "affected_entity")]
    non_protective = data.loc[~data.index.isin(protective_causes, "affected_entity")]

    if not protective.empty:
        check_value_columns_boundary(
            protective,
            boundary_value=VALID_PROTECTIVE_PAF_MIN,
            boundary_type="lower",
            value_columns=DRAW_COLUMNS,
            error=DataTransformationError,
        )
        check_value_columns_boundary(
            protective,
            boundary_value=VALID_PAF_RANGE[0],
            boundary_type="upper",
            value_columns=DRAW_COLUMNS,
        )
        check_value_columns_boundary(
            protective,
            boundary_value=VALID_PAF_RANGE[1],
            boundary_type="upper",
            value_columns=DRAW_COLUMNS,
            error=DataTransformationError,
        )
    if not non_protective.empty:
        error = None if entity.name in RISKS_WITH_NEGATIVE_PAF else DataTransformationError
        check_value_columns_boundary(
            non_protective,
            boundary_value=VALID_PAF_RANGE[0],
            boundary_type="lower",
            value_columns=DRAW_COLUMNS,
            error=error,
        )
        check_value_columns_boundary(
            non_protective,
            boundary_value=VALID_PAF_RANGE[1],
            boundary_type="upper",
            value_columns=DRAW_COLUMNS,
            error=DataTransformationError,
        )

    for (c_name, measure), g in risk_relationship:
        cause = [c for c in causes if c.name == c_name][0]
        if measure == "incidence_rate":
            check_age_restrictions(g, cause, rest_type="yld", fill_value=0.0, context=context)
        else:  # excess mortality
            check_age_restrictions(g, cause, rest_type="yll", fill_value=0.0, context=context)
        check_sex_restrictions(
            g, cause.restrictions.male_only, cause.restrictions.female_only, fill_value=0.0
        )


def validate_mediation_factors(
    data: pd.DataFrame,
    entity: RiskFactor,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    raise NotImplementedError()


def validate_estimate(
    data: pd.DataFrame,
    entity: Covariate,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped covariate
    estimate data, adjusting as needed for covariates that are not by age or
    by sex to skip those column checks. Check that the lower, mean, and upper
    values for each demographic group are either all 0 or lower < mean < upper.

    Parameters
    ----------
    data
        Simulation-prepped covariate estimate data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yld age or sex restrictions are violated, or data does not satisfy
        lower = mean = upper = 0 or lower < mean < upper for each demographic
        group.

    """
    expected_index_names = ["location", "year", "parameter"]
    if entity.by_sex:
        validate_sex_column(data)
        expected_index_names += ["sex"]
    if entity.by_age:
        validate_age_column(data, context)
        expected_index_names += ["age"]

    validate_expected_index_and_columns(
        expected_index_names, data.index.names, ["value"], data.columns
    )
    validate_location_column(data, context)
    validate_year_column(data, context)
    validate_value_column(data)

    cols = list(set(expected_index_names).difference({"parameter", "value"}))
    data.groupby(cols).apply(check_covariate_values)


def validate_cost(
    data: pd.DataFrame,
    entity: HealthTechnology | HealthcareEntity,
    context: SimulationValidationContext,
) -> None:
    """Check the standard set of validations on simulation-prepped cost data.

    Parameters
    ----------
    data
        Simulation-prepped cost data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yld age or sex restrictions are violated, or data falls outside the
        expected boundary values.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS + [entity.kind]
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    validate_standard_columns(data, context)
    check_value_columns_boundary(
        data,
        VALID_COST_RANGE[0],
        "lower",
        value_columns=DRAW_COLUMNS,
        inclusive=True,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        VALID_COST_RANGE[1][entity.kind],
        "upper",
        value_columns=DRAW_COLUMNS,
        inclusive=True,
        error=DataTransformationError,
    )


def validate_utilization_rate(
    data: pd.DataFrame, entity: HealthcareEntity, context: SimulationValidationContext
) -> None:
    """Check the standard set of validations on simulation-prepped utilization
    data.

    Parameters
    ----------
    data
        Simulation-prepped cost utilization to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yld age or sex restrictions are violated, or data falls outside the
        expected boundary values.

    """
    expected_index_names = SCRUBBED_DEMOGRAPHIC_COLUMNS
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, DRAW_COLUMNS, data.columns
    )

    validate_standard_columns(data, context)
    check_value_columns_boundary(
        data,
        VALID_UTILIZATION_RANGE[0],
        "lower",
        value_columns=DRAW_COLUMNS,
        inclusive=True,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        VALID_UTILIZATION_RANGE[1],
        "upper",
        value_columns=DRAW_COLUMNS,
        inclusive=True,
        error=DataTransformationError,
    )


def validate_structure(
    data: pd.DataFrame,
    entity: Population,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Check the standard set of validations on simulation-prepped population
    structure data, skipping the check on the draw column since structure data
    isn't by draw.

    Parameters
    ----------
    data
        Simulation-prepped population structure data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any standard columns are incorrectly named or contain invalid values,
        if yld age or sex restrictions are violated, or data falls outside the
        expected boundary values.

    """
    validate_demographic_columns(data, context)
    validate_value_column(data)

    check_value_columns_boundary(
        data,
        VALID_POPULATION_RANGE[0],
        "lower",
        value_columns=["value"],
        inclusive=True,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        VALID_POPULATION_RANGE[1],
        "upper",
        value_columns=["value"],
        inclusive=True,
        error=DataTransformationError,
    )


def validate_theoretical_minimum_risk_life_expectancy(
    data: pd.DataFrame,
    entity: Population,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """Because the structure of life expectancy data is somewhat different,
    containing a custom age column that doesn't match the standard GBD age
    groups, this validator doesn't use the standard column checks. Instead, it
    verifies that the data has the correct age columns and that ages range from
    0 to 110 years. It checks that all life expectancy values are within the
    expected range.

    Parameters
    ----------
    data
        Simulation-prepped theoretical minimum risk life expectancy data.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If age columns are incorrectly named or contain invalid values or if
        any life expectancy values are outside the expected range.

    """
    expected_index_names = ["age"]
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, ["value"], data.columns
    )

    validate_value_column(data)

    age_min, age_max = 0, 110
    if data.index.left.min() > age_min or data.index.right.max() < age_max:
        raise DataTransformationError(
            f"Life expectancy data does not span the "
            f"entire age range [{age_min}, {age_max}]."
        )

    check_value_columns_boundary(
        data,
        VALID_LIFE_EXP_RANGE[0],
        "lower",
        value_columns=["value"],
        inclusive=False,
        error=DataTransformationError,
    )
    check_value_columns_boundary(
        data,
        VALID_LIFE_EXP_RANGE[1],
        "upper",
        value_columns=["value"],
        inclusive=False,
        error=DataTransformationError,
    )


def validate_age_bins(
    data: pd.DataFrame,
    entity: Population,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """With only age columns in this data, the validator is an abbreviated
    version employing only the standard column check on ages.

    Parameters
    ----------
    data
        Simulation-prepped age bin data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any age columns are incorrectly named or contain invalid values.

    """
    expected_index_names = ["age", "age_group_name"]
    validate_expected_index_and_columns(
        expected_index_names, data.index.names, [], data.columns
    )

    validate_age_column(data, context=context)


def validate_demographic_dimensions(
    data: pd.DataFrame,
    entity: Population,
    context: SimulationValidationContext,
    value_columns: list[str],
) -> None:
    """With only demographic columns in this data, the validator is an
    abbreviated version employing only the standard demographic column checks.

    Parameters
    ----------
    data
        Simulation-prepped demographic dimensions data to validate.
    entity
        Entity to which the data pertain.
    context
        Wrapper for additional data used in the validation process.
    value_columns
        List of column names in `data` that contain the values to be validated.

    Raises
    ------
    DataTransformationError
        If any demographic columns are incorrectly named or contain invalid
        values.

    """
    validate_demographic_columns(data, context)


#############
# UTILITIES #
#############


def validate_expected_index_and_columns(
    expected_index_names: list,
    existing_index_names: list,
    expected_cols: list,
    existing_cols: list,
) -> None:
    """Verify that the passed lists of columns and index names match.

    Parameters
    ----------
    expected_index_names
        List of index names expected.
    existing_index_names
        List of index names actually found in data.
    expected_cols
        List of column names expected.
    existing_cols
        List of column names actually found in data.

    Raises
    ------
    DataTransformationError
        If `expected_index_names` doesn't match of `existing_index_names` or
        `expected_cols` does not match `existing_cols`.
    """
    to_check = [
        (set(expected_index_names), set(existing_index_names), "index names"),
        (set(expected_cols), set(existing_cols), "columns"),
    ]
    for c in to_check:
        if c[1] < c[0]:
            raise DataTransformationError(
                f"Data is missing {c[2]}: " f"{c[0].difference(c[1])}."
            )
        elif c[1] > c[0]:
            raise DataTransformationError(
                f"Data returned extra {c[2]} " f"{c[1].difference(c[0])}."
            )


def validate_standard_columns(
    data: pd.DataFrame, context: SimulationValidationContext
) -> None:
    """Validate that dataframe have expected column names and values.

    The dataframe is expected to have location, sex, age, year, and value columns.

    Parameters
    ----------
    data
        Simulation-prepped data to validate.
    context
        Wrapper for additional data used in the validation process.

    Raises
    ------
    DataTransformationError
        If any location, sex, age, year, draw, or value columns are incorrectly
        named or contain invalid values.
    """
    validate_demographic_columns(data, context)
    validate_value_column(data)


def validate_demographic_columns(
    data: pd.DataFrame, context: SimulationValidationContext
) -> None:
    """Validate that demographic (location, sex, age, and year) columns in the
    passed dataframe all have the expected names and values. The given context
    provides the full range of expected values for location, age, and year.

    Parameters
    ----------
    data
        Simulation-prepped data to validate.
    context
        Wrapper for additional data used in the validation process.

    Raises
    ------
    DataTransformationError
        If any demographic columns are incorrectly named or contain invalid
        values.

    """
    validate_location_column(data, context)
    validate_sex_column(data)
    validate_age_column(data, context=context)
    validate_year_column(data, context)


def validate_location_column(
    data: pd.DataFrame, context: SimulationValidationContext
) -> None:
    """Validate that location index column in the data has the expected value.

    Parameters
    ----------
    data
        Simulation-prepped data to validate.
    context
        Wrapper for additional data used in the validation process

    Raises
    ------
    DataTransformationError
        If 'location' column does not contain only the single location given
        in `context`.

    """
    data_locations = data.index.unique("location")
    if not set(data_locations) == (set(context["location"])):
        # Locations requested for extraction not found in data
        missing_locations_in_data = set(context["location"]).difference(data_locations)
        if missing_locations_in_data:
            missing_error_message = (
                f"Locations missing in data include '{missing_locations_in_data}'. "
            )
        # Locations found in data but not requested for extraction
        extra_locations_in_data = set(data_locations).difference(context["location"])
        if extra_locations_in_data:
            extra_error_message = (
                f"Extra locations found in data include '{extra_locations_in_data}'. "
            )
        raise DataTransformationError(
            "Location(s) must match between data and SimulationValidationContext. "
            f"{missing_error_message} + {extra_error_message}"
        )


def validate_sex_column(data: pd.DataFrame) -> None:
    """Validate that sex index column in the data has the expected values.

    Parameters
    ----------
    data
        Simulation-prepped data to validate.

    Raises
    ------
    DataTransformationError
        If 'sex' column does not contain only the values 'Male' and 'Female'.

    """
    if set(data.index.unique("sex")) != {"Male", "Female"}:
        raise DataTransformationError(
            "Sex must contain 'Male' and 'Female' and nothing else."
        )


def validate_age_column(data: pd.DataFrame, context: SimulationValidationContext) -> None:
    """Validate that age index column in the data has the expected values.

    Parameters
    ----------
    data
        Simulation-prepped data to validate.
    context
        Wrapper for additional data used in validation.

    Raises
    ------
    DataTransformationError
        If 'age' index does not contain the full range of expected
        age bins supplied in `context`.

    """
    expected_ages = [
        pd.Interval(row.age_start, row.age_end, closed="left")
        for _, row in context["age_bins"].iterrows()
    ]
    data_ages = data.index.levels[data.index.names.index("age")]

    if not sorted(data_ages) == sorted(expected_ages):
        raise DataTransformationError(
            "Age_start and age_end must contain all gbd age groups."
        )


def validate_year_column(data: pd.DataFrame, context: SimulationValidationContext) -> None:
    """Validate that year column in the data has the expected values.

    Parameters
    ----------
    data
        Simulation-prepped data to validate.
    context
        Wrapper for additional data used in validation.

    Raises
    ------
    DataTransformationError
        If 'year' index column does not contain the full range of expected year bins
        supplied in `context`.

    """
    expected_years = [
        pd.Interval(row.year_start, row.year_end, closed="left")
        for _, row in context["years"].iterrows()
    ]
    data_years = data.index.levels[data.index.names.index("year")]

    if not sorted(data_years) == sorted(expected_years):
        formatted_expected_years = [interval.left for interval in sorted(expected_years)]
        formatted_data_years = [interval.left for interval in sorted(data_years)]
        raise DataTransformationError(
            f"Data was expected to contain years {formatted_expected_years}. The data provided contains years {formatted_data_years}."
        )


def validate_value_column(data: pd.DataFrame) -> None:
    """Validate that value columns (i.e., any non-index columns) have no missingness.

    Parameters
    ----------
    data
        Simulation-prepped data to validate.

    Raises
    ------
    DataTransformationError
        If any non-index columns contain any NaN or Inf values.
    """

    if np.any(np.isnan(data.values)):
        raise DataTransformationError("Value data found to contain NaN.")
    if np.any(np.isinf(data.values)):
        raise DataTransformationError("Value data found to contain infinity.")


def check_age_restrictions(
    data: pd.DataFrame,
    entity: ModelableEntity,
    rest_type: str,
    fill_value: float | dict[str, float],
    context: SimulationValidationContext,
):
    """Given an entity and which restrictions to use, ensure that all data for
    age groups outside of the range of restrictions for that entity has values
    only of fill_value.

    Parameters
    ----------
    data
        Simulation-prepped data to validate.
    entity
        Entity for which to validate restrictions.
    rest_type
        Which restrictions from entity to use. One of: yll, yld, inner, outer.
    fill_value
        The only allowable value in data outside of age restrictions. For
        categorical risks, this should be dictionary containing values for
        'exposed' and 'unexposed' categories.
    context
        Wrapper containing additional data used in the simulation.

    Raises
    ------
    DataTransformationError
        If any values other than fill_value are found in any non-index columns
        in data outside the restrictions of entity.

    """
    start_id, end_id = utilities.get_age_group_ids_by_restriction(entity, rest_type)

    age_bins = context["age_bins"]
    id_to_age_start_map = dict(zip(age_bins.age_group_id, age_bins.age_start))
    age_range_start = id_to_age_start_map[start_id]
    age_range_end = id_to_age_start_map[end_id]
    in_range_ages = age_bins.loc[
        (age_bins.age_start >= age_range_start) & (age_bins.age_start <= age_range_end)
    ]
    in_range_age_intervals = [
        pd.Interval(row.age_start, row.age_end, closed="left")
        for _, row in in_range_ages.iterrows()
    ]
    outside = data.loc[~data.index.isin(in_range_age_intervals, "age")]

    if (
        entity.kind in ["risk_factor", "alternative_risk_factor"]
        and entity.distribution
        in ["dichotomous", "ordered_polytomous", "unordered_polytomous"]
        and isinstance(fill_value, dict)
    ):
        _check_cat_risk_fill_values(outside, entity, fill_value, "age")

    elif not outside.empty and np.any(outside.values != fill_value):
        raise DataTransformationError(
            f"Age restrictions are violated by a value other than fill={fill_value}."
        )


def check_sex_restrictions(
    data: pd.DataFrame,
    male_only: bool,
    female_only: bool,
    fill_value: float | dict[str, float],
    entity=None,
):
    """Given an entity and which restrictions to use, ensure that all data for
    sexes outside of the restrictions for that entity has values only of
    fill_value.
    Parameters
    ----------
    data
        Simulation-prepped data to validate.
    female_only
        Boolean indicating whether the data should be restricted to females
        only. If true, all male data should have values only of fill_value.
    male_only
        Boolean indicating whether the data should be restricted to females
        only. If true, all female data should have values only of fill_value.
    fill_value
        The only allowable value in data outside of age restrictions. For
        categorical risks, this should be dictionary containing values for
        'exposed' and 'unexposed' categories.
    context
        Wrapper containing additional data used in the simulation.
    entity
        Optional. Used to check if the entity is a categorical risk, in which
        case only used for exposure data validation where the fill_values vary
        by category.
    Raises
    ------
    DataTransformationError
        If any values other than fill_value are found in any non-index columns
        in data outside the restrictions of entity.
    """
    outside = None
    if male_only:
        outside = data.xs("Female", level="sex")
        sex = "male"
    elif female_only:
        outside = data.xs("Male", level="sex")
        sex = "female"
    if outside is not None:
        if entity is not None and (
            entity.kind in ["risk_factor", "alternative_risk_factor"]
            and entity.distribution
            in ["dichotomous", "ordered_polytomous", "unordered_polytomous"]
            and isinstance(fill_value, dict)
        ):
            _check_cat_risk_fill_values(outside, entity, fill_value, "sex")

        elif np.any(outside.values != fill_value):
            raise DataTransformationError(
                f"Restriction to {sex} sex only is violated "
                f"by a value other than fill={fill_value}."
            )


def _check_cat_risk_fill_values(
    outside_data: pd.DataFrame,
    entity: RiskFactor | AlternativeRiskFactor,
    fill_value: dict[str, float],
    restriction: str,
):
    """Helper method for checking restrictions for categorical risks where two
    fill values are allowed: one for exposed categories and one for the
    unexposed category.

    Parameters
    ----------
    outside_data
        Simulation-prepped data outside the restrictions.
    entity
        Entity to which the data pertain.
    fill_value
        Dictionary containing the fill values for the exposed and unexposed
        categories.
    restriction
        Whether the restriction being checked is 'sex' or 'age'.

    Raises
    ------
    DataTransformationError
        If the outside_data contains values other than fill_value in any
        non-index columns for the correct categories.

    """
    tmrel_cat = utility_data.get_tmrel_category(entity)
    outside_unexposed = outside_data.loc[
        outside_data.index.get_level_values("parameter") == tmrel_cat
    ]
    outside_exposed = outside_data.loc[
        outside_data.index.get_level_values("parameter") != tmrel_cat
    ]
    if not outside_unexposed.empty and np.any(
        outside_unexposed.values != fill_value["unexposed"]
    ):
        raise DataTransformationError(
            f"{restriction.capitalize()} restrictions for TMREL cat are violated by a "
            f'value other than fill={fill_value["unexposed"]}.'
        )
    if not outside_exposed.empty and np.any(outside_exposed.values != fill_value["exposed"]):
        raise DataTransformationError(
            f"{restriction.capitalize()} restrictions for non-TMREL categories are violated "
            f'by a value other than fill={fill_value["exposed"]}.'
        )


def check_covariate_values(data: pd.DataFrame) -> None:
    """Validator for covariate estimate data to check that for each demographic
    group either lower, mean, and upper values are all 0 or
    lower < mean < upper.

    Parameters
    ----------
    data
        Simulation-prepped covariate estimate data for a single demographic
        group.

    Raises
    ------
    DataTransformationError
        If lower, mean, and upper values are not all 0 and it is not the case
         that lower < mean < upper.
    """
    lower = data.xs("lower_value", level="parameter").value.values
    mean = data.xs("mean_value", level="parameter").value.values
    upper = data.xs("upper_value", level="parameter").value.values

    # allow the case where lower = mean = upper = 0 b/c of things like age
    # specific fertility rate where all estimates are 0 for young age groups
    if np.all(data.value != 0) and not np.all(lower <= mean <= upper):
        raise DataTransformationError(
            "Covariate data contains demographic groups for which the "
            "estimates for lower, mean, and upper values are not all 0 "
            "and it is not the case that lower <= mean <= upper. "
        )



import re

import pandas as pd
import pytest

from vivarium_inputs import utilities
from vivarium_inputs.globals import (
    DRAW_COLUMNS,
    MEAN_COLUMNS,
    NON_STANDARD_MEASURES,
    SUPPORTED_DATA_TYPES,
)

POSSIBLE_MEASURES = [
    "incidence_rate",
    "raw_incidence_rate",
    "prevalence",
    "birth_prevalence",
    "disability_weight",
    "remission_rate",
    "cause_specific_mortality_rate",
    "excess_mortality_rate",
    "deaths",
    # Risk-like measures
    "exposure",
    "exposure_standard_deviation",
    "exposure_distribution_weights",
    "relative_risk",
    "population_attributable_fraction",
    # Covariate measures
    "estimate",
    # Population measures
    "structure",
    "theoretical_minimum_risk_life_expectancy",
    "age_bins",
    "demographic_dimensions",
]


@pytest.mark.parametrize(
    "sex_ids",
    [(1, 1, 1, 2, 2, 2), (1, 1, 2, 2, 3, 3), (1, 1, 1), (2, 2, 2), (3, 3, 3)],
    ids=["male_female", "male_female_both", "male", "female", "both"],
)
def test_normalize_sex(sex_ids):
    df = pd.DataFrame({"sex_id": sex_ids, "value": [1] * len(sex_ids)})
    normalized = utilities.normalize_sex(df, fill_value=0.0, cols_to_fill=["value"])
    assert {1, 2} == set(normalized.sex_id)


def test_normalize_sex_copy_3():
    values = [1, 2, 3, 4]
    df = pd.DataFrame({"sex_id": [3] * len(values), "value": values})
    normalized = utilities.normalize_sex(df, fill_value=0.0, cols_to_fill=["value"])
    assert (normalized.loc[normalized.sex_id == 1, "value"] == values).all()
    assert (normalized.loc[normalized.sex_id == 2, "value"] == values).all()


def test_normalize_sex_fill_value():
    values = [1, 2, 3, 4]
    fill = 0.0
    for sex in [1, 2]:
        missing_sex = 1 if sex == 2 else 2
        df = pd.DataFrame({"sex_id": [sex] * len(values), "value": values})
        normalized = utilities.normalize_sex(df, fill_value=fill, cols_to_fill=["value"])
        assert (normalized.loc[normalized.sex_id == sex, "value"] == values).all()
        assert (
            normalized.loc[normalized.sex_id == missing_sex, "value"] == [fill] * len(values)
        ).all()


def test_normalize_sex_no_sex_id():
    df = pd.DataFrame({"ColumnA": [1, 2, 3], "ColumnB": [1, 2, 3]})
    normalized = utilities.normalize_sex(df, fill_value=0.0, cols_to_fill=["value"])
    pd.testing.assert_frame_equal(df, normalized)


@pytest.mark.parametrize(
    "data_type, should_raise",
    [
        ("means", False),
        ("draws", False),
        ("foo", True),
        (["means", "draws"], True),  # temporarily; not implemented
        (["means", "draws", "foo"], True),
        ({"not": "a list"}, True),
    ],
)
def test_validate_data_type(data_type, should_raise):
    if should_raise:
        if not isinstance(data_type, (list, str)):
            match = re.escape(
                f"'data_type' must be a string or a list of strings. Got {type(data_type)}."
            )
        elif isinstance(data_type, list):
            match = "Lists of data types are not yet supported."
        else:
            match = re.escape(
                f"Data type(s) {set(['foo'])} are not supported. Supported types are {list(SUPPORTED_DATA_TYPES)}."
            )
        with pytest.raises(utilities.DataTypeNotImplementedError, match=match):
            utilities.DataType("foo", data_type)
    else:
        utilities.DataType("foo", data_type)


@pytest.mark.parametrize("measure", POSSIBLE_MEASURES, ids=lambda x: x)
@pytest.mark.parametrize(
    "data_type",
    ("means", "draws", None, ["means", "draws"]),
    ids=["means", "draws", "None", "means_draws"],
)
def test_get_value_columns(measure, data_type):
    if isinstance(data_type, list):
        with pytest.raises(
            utilities.DataTypeNotImplementedError,
            match="Lists of data types are not yet supported.",
        ):
            utilities.DataType(measure, data_type).value_columns
    else:
        if data_type == None:  # Hacky: this goes first in the DataType class
            expected_returned_cols = []
        elif measure in NON_STANDARD_MEASURES:
            expected_returned_cols = ["value"]
        elif data_type == "means":
            expected_returned_cols = MEAN_COLUMNS
        else:  # draws
            expected_returned_cols = DRAW_COLUMNS
        assert utilities.DataType(measure, data_type).value_columns == expected_returned_cols



from gbd_mapping.id import me_id

from .healthcare_entity_template import HealthcareEntities, HealthcareEntity

healthcare_entities = HealthcareEntities(
    outpatient_visits=HealthcareEntity(
        name="outpatient_visits",
        kind="healthcare_entity",
        gbd_id=me_id(19797),
        utilization=me_id(19797),
    ),
    inpatient_visits=HealthcareEntity(
        name="inpatient_visits",
        kind="healthcare_entity",
        gbd_id=me_id(18749),
        utilization=me_id(18749),
    ),
)



import pandas as pd
from gbd_mapping import RiskFactor

from vivarium_inputs.globals import NON_MAX_TMREL, NUM_DRAWS, SEXES, gbd


def get_estimation_years(*_, **__) -> pd.Series:
    return gbd.get_estimation_years()


def get_most_recent_year(*_, **__) -> int:
    return gbd.get_most_recent_year()


def get_year_block(*_, **__) -> pd.DataFrame:
    estimation_years = get_estimation_years()
    year_block = pd.DataFrame(
        {"year_start": range(min(estimation_years), max(estimation_years) + 1)}
    )
    year_block["year_end"] = year_block["year_start"] + 1
    return year_block


def get_age_group_ids(*_, **__) -> list[int]:
    return gbd.get_age_group_id()


def get_age_bins(*_, **__) -> pd.DataFrame:
    age_bins = get_raw_age_bins()[
        ["age_group_id", "age_group_name", "age_group_years_start", "age_group_years_end"]
    ].rename(columns={"age_group_years_start": "age_start", "age_group_years_end": "age_end"})
    age_bins = age_bins.sort_values("age_start").reset_index(drop=True)
    return age_bins


def get_raw_age_bins() -> pd.DataFrame:
    """Wrapper for gbd call for test mocking purposes."""
    return gbd.get_age_bins()


def get_location_id(location_name: str) -> int:
    return {r.location_name: r.location_id for _, r in get_raw_location_ids().iterrows()}[
        location_name
    ]


def get_location_name(location_id: int) -> str:
    return {
        row.location_id: row.location_name for _, row in get_raw_location_ids().iterrows()
    }[location_id]


def get_raw_location_ids() -> pd.DataFrame:
    """Wrapper for gbd call for test mocking purposes."""
    return gbd.get_location_ids()


def get_location_id_parents(location_id: int | list[int]) -> dict[int, list]:
    if isinstance(location_id, int):
        location_id = [location_id]
    location_metadata = get_location_path_to_global()
    parent_ids = (
        location_metadata.loc[location_metadata.location_id.isin(location_id)]
        .set_index("location_id")["path_to_top_parent"]
        .str.split(",")
        .to_dict()
    )
    # Coerce list of parent ids to integers
    parent_ids = {loc_id: list(map(int, parents)) for loc_id, parents in parent_ids.items()}

    return parent_ids


def get_location_path_to_global() -> pd.DataFrame:
    """Wrapper for gbd call for test mocking purposes."""
    return gbd.get_location_path_to_global()


def get_demographic_dimensions(
    location_id: int | list[int],
    draws: bool = False,
    value: float = None,
    years: int | str | list[int] | None = None,
) -> pd.DataFrame:
    ages = get_age_group_ids()
    estimation_years = get_estimation_years()
    if years is None:  # default to most recent year
        years = [get_most_recent_year()]
    elif years == "all":
        years = range(min(estimation_years), max(estimation_years) + 1)
    else:
        years = years if isinstance(years, list) else [years]
        not_estimated_years = set(years).difference(estimation_years)
        if not_estimated_years:
            raise ValueError(
                f"years must be in {estimation_years}. You provided {not_estimated_years}."
            )

    sexes = [SEXES["Male"], SEXES["Female"]]
    location = [location_id] if isinstance(location_id, int) else location_id
    values = [location, sexes, ages, years]
    names = ["location_id", "sex_id", "age_group_id", "year_id"]

    data = pd.MultiIndex.from_product(values, names=names).to_frame(index=False)
    if draws:
        for i in range(NUM_DRAWS):
            data[f"draw_{i}"] = value
    return data


def get_tmrel_category(entity: RiskFactor) -> str:
    if entity.name in NON_MAX_TMREL:
        tmrel_cat = NON_MAX_TMREL[entity.name]
    else:
        tmrel_cat = sorted(list(entity.categories.to_dict()), key=lambda x: int(x[3:]))[-1]
    return tmrel_cat



from gbd_mapping.base_template import GbdRecord
from gbd_mapping.risk_factor_template import RiskFactor


class AlternativeRiskFactor(RiskFactor):
    """Container for alternative GBD risk factor definitions."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AlternativeRiskFactors(GbdRecord):
    """Holder of alternative risk factors."""

    __slots__ = (
        "child_wasting",
        "child_underweight",
        "child_stunting",
    )

    def __init__(
        self,
        child_stunting: AlternativeRiskFactor,
        child_underweight: AlternativeRiskFactor,
        child_wasting: AlternativeRiskFactor,
    ):
        self.child_stunting = child_stunting
        self.child_underweight = child_underweight
        self.child_wasting = child_wasting



from __future__ import annotations

import logging
from typing import Any

import pytest
from _pytest.logging import caplog as _caplog
from loguru import logger


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture
def runslow(request: pytest.FixtureRequest):
    """Fixture to allow 'runslow' to be used as an argument in tests."""
    return request.config.getoption("--runslow")


@pytest.fixture
def caplog(_caplog):
    class PropogateHandler(logging.Handler):
        def emit(self, record):
            logging.getLogger(record.name).handle(record)

    handler_id = logger.add(PropogateHandler(), format="{message}")
    yield _caplog
    logger.remove(handler_id)


def pytest_report_teststatus(report: Any, config: Any) -> tuple[Any, str, str] | None:
    if report.when == "call" and report.skipped:
        if "todo" in report.longreprtext.lower():
            return (report.outcome, "T", "TODO")
        if "fixme" in report.longreprtext.lower():
            # NOTE: We mark the non-verbose status as "T" (the same as a TODO) because
            # "F" is reserved for failures.
            return (report.outcome, "T", "FIXME")


def _no_gbd_access():
    try:
        from vivarium_inputs.globals import GBDDummy

        return True
    except ImportError:
        pass
    return False


NO_GBD_ACCESS = _no_gbd_access()


def is_not_implemented(data_type: str | list[str], measure: str) -> bool:
    return isinstance(data_type, list) or (
        data_type == "means"
        and measure
        in [
            "disability_weight",
            "remission_rate",
            "cause_specific_mortality_rate",
            "excess_mortality_rate",
            "deaths",
            "exposure",
            "exposure_standard_deviation",
            "exposure_distribution_weights",
            "relative_risk",
            "population_attributable_fraction",
            "estimate",
            "structure",
        ]
    )



import operator

import numpy as np
import pandas as pd
from loguru import logger

from vivarium_inputs.globals import VivariumInputsError

###############################
# Shared validation utilities #
###############################


def check_value_columns_boundary(
    data: pd.DataFrame,
    boundary_value: float | pd.Series,
    boundary_type: str,
    value_columns: list[str],
    inclusive: bool = True,
    error: type(VivariumInputsError) = None,
):
    """Check that all values in DRAW_COLUMNS in data are above or below given
    boundary_value.

    Parameters
    ----------
    data
        Dataframe containing `value_columns`.
    boundary_value
        Value against which `value_columns` values will be checked. May be a
        series of values with a matching index to data.
    boundary_type
        String 'upper' or 'lower' indicating whether `boundary_value` is upper
        or lower limit on `value_columns`.
    value_columns
        List of column names in `data`, the values of which should be checked
        against `boundary_value`.
    inclusive
        Boolean indicating whether `boundary_value` is inclusive or not.
    error
        Exception class indicating what error should be raised if values are
        found outside `boundary_value`. If none, warn instead of raising error.

    Raises
    -------
    ValueError
        If the passed `boundary_type` is neither 'upper' nor 'lower'.
    error
        If any values in `value_columns` are above/below `boundary_value`,
        depending on `boundary_type`, raise the passed error if there is one.
    """
    msg = (
        f'Data contains values {"below" if boundary_type == "lower" else "above"} '
        f'{"or equal to " if not inclusive else ""}the expected boundary '
        f'value{"s" if isinstance(boundary_value, pd.Series) else f" ({boundary_value})"}.'
    )

    if boundary_type == "lower":
        op = operator.gt
        data_values = data[value_columns].min(axis=1)
    elif boundary_type == "upper":
        op = operator.lt
        data_values = data[value_columns].max(axis=1)
    else:
        raise ValueError(
            f'Boundary type must be either "lower" or "upper". You specified {boundary_type}.'
        )

    if isinstance(boundary_value, pd.Series):
        data_values = data_values.sort_index()
        boundary_value = boundary_value.sort_index()

    within_boundary = op(data_values, boundary_value)
    if inclusive:
        within_boundary |= np.isclose(data_values, boundary_value)

    if not np.all(within_boundary):
        if error is not None:
            raise error(msg)
        else:
            logger.warning(msg)



class DummyLoadComponent:
    def __init__(self, entity_key: str):
        self.entity_key = entity_key

    @property
    def name(self):
        return f"dummy_load_component({self.entity_key})"

    def setup(self, builder):
        builder.data.load(self.entity_key)
        print(f"data loaded for {self.entity_key}")



from .alternative_risk_factor import AlternativeRiskFactor, alternative_risk_factors
from .health_technology import HealthTechnology, health_technologies
from .healthcare_entity import HealthcareEntity, healthcare_entities



from __future__ import annotations

import numpy as np
import pandas as pd
from gbd_mapping import Cause, Covariate, Etiology, ModelableEntity, RiskFactor, Sequela

import vivarium_inputs.validation.raw as validation
from vivarium_inputs import utility_data
from vivarium_inputs.globals import (
    DRAW_COLUMNS,
    MEASURES,
    METRICS,
    OTHER_MEID,
    DataAbnormalError,
    DataDoesNotExistError,
    EmptyDataFrameException,
    InputsException,
    NoBestVersionsException,
    Population,
    gbd,
)
from vivarium_inputs.mapping_extension import AlternativeRiskFactor, HealthcareEntity
from vivarium_inputs.utilities import (
    DataType,
    filter_to_most_detailed_causes,
    process_kidney_dysfunction_exposure,
)


def extract_data(
    entity: ModelableEntity,
    measure: str,
    location_id: list[int],
    years: int | str | list[int] | None,
    data_type: DataType,
    validate: bool = True,
) -> pd.Series | pd.DataFrame:
    """Pull raw data from GBD.

    Also checks metadata for the requested entity-measure pair.
    The only filtering that occurs is by applicable measure id, metric id,
    or to most detailed causes where relevant. If validate is turned on, will
    also pull any additional data needed for raw validation and call raw
    validation on the extracted data.

    Parameters
    ----------
    entity
        Entity for which to extract data.
    measure
        Measure for which to extract data.
    location_id
        Location for which to extract data.
    years
        Years for which to extract data. If None, get most recent year. If 'all',
        get all available data.
    data_type
        DataType object for which to extract data.
    validate
        Flag indicating whether additional data needed for raw validation
        should be extracted and whether raw validation should be performed.
        Should only be set to False if data is being extracted for
        investigation. Never extract data for a simulation without validation.
        Defaults to True.

    Returns
    -------
        Data for the entity-measure pair and specific location requested.

    Raises
    ------
    DataDoesNotExistError
        If data for the entity-measure-location set requested does not exist.

    """
    extractors = {
        # Cause-like measures
        "incidence_rate": (extract_incidence_rate, {}),
        "prevalence": (extract_prevalence, {}),
        "birth_prevalence": (extract_birth_prevalence, {}),
        "disability_weight": (extract_disability_weight, {}),
        "remission_rate": (extract_remission_rate, {}),
        "deaths": (extract_deaths, {"population": extract_structure}),
        # Risk-like measures
        "exposure": (extract_exposure, {}),
        "exposure_standard_deviation": (
            extract_exposure_standard_deviation,
            {"exposure": extract_exposure},
        ),
        "exposure_distribution_weights": (extract_exposure_distribution_weights, {}),
        "relative_risk": (extract_relative_risk, {"exposure": extract_exposure}),
        "population_attributable_fraction": (
            extract_population_attributable_fraction,
            {"exposure": extract_exposure, "relative_risk": extract_relative_risk},
        ),
        "etiology_population_attributable_fraction": (
            extract_population_attributable_fraction,
            {},
        ),
        "mediation_factors": (extract_mediation_factors, {}),
        # Covariate measures
        "estimate": (extract_estimate, {}),
        # Health system measures
        "utilization_rate": (extract_utilization_rate, {}),
        # Population measures
        "structure": (extract_structure, {}),
        "theoretical_minimum_risk_life_expectancy": (
            extract_theoretical_minimum_risk_life_expectancy,
            {},
        ),
    }

    validation.check_metadata(entity, measure)

    year_id = _get_year_id(years, data_type)

    try:
        main_extractor, additional_extractors = extractors[measure]
        data = main_extractor(entity, location_id, year_id, data_type)
    except (
        ValueError,
        AssertionError,
        EmptyDataFrameException,
        NoBestVersionsException,
        InputsException,
    ) as e:
        if isinstance(
            e, ValueError
        ) and f"Metadata associated with rei_id = {entity.gbd_id}" not in str(e):
            raise e
        elif (
            isinstance(e, AssertionError)
            and f"Invalid covariate_id {entity.gbd_id}" not in str(e)
            and "No best model found" not in str(e)
        ):
            raise e
        elif isinstance(e, InputsException) and measure != "birth_prevalence":
            raise e
        else:
            raise DataDoesNotExistError(
                f"{measure.capitalize()} data for {entity.name} does not exist."
            )

    # drop extra draw columns
    existing_draw_cols = [col for col in data if col.startswith("draw_")]
    extra_draw_cols = [col for col in existing_draw_cols if col not in DRAW_COLUMNS]
    data = data.drop(columns=extra_draw_cols, errors="ignore")

    # drop get_outputs data earlier than the estimation years
    if data_type.type == "means":
        data = data.loc[data["year_id"] >= min(utility_data.get_estimation_years())]

    if validate:
        additional_data = {
            name: extractor(entity, location_id, year_id, data_type)
            for name, extractor in additional_extractors.items()
        }
        if year_id:  # if not pulling all years
            additional_data["estimation_years"] = (
                [year_id] if not isinstance(year_id, list) else year_id
            )
        validation.validate_raw_data(
            data, entity, measure, location_id, data_type.value_columns, **additional_data
        )

    return data


####################
# Helper functions #
####################


def _get_year_id(years, data_type):
    if years is None:  # default to most recent year
        year_id = utility_data.get_most_recent_year()
    elif years == "all":
        if data_type.type == "draws":
            year_id = None
        else:  # means
            year_id = "full"
    else:
        year_id = years if isinstance(years, list) else [years]
        estimation_years = utility_data.get_estimation_years()
        not_estimated_years = set(year_id).difference(estimation_years)
        if not_estimated_years:
            raise ValueError(
                f"years must be in {estimation_years}. You provided {not_estimated_years}."
            )

    return year_id


#######################
# Extractor functions #
#######################


def extract_incidence_rate(
    entity: Cause | Sequela,
    location_id: list[int],
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:

    data = gbd.get_incidence_prevalence(
        entity_id=entity.gbd_id,
        location_id=location_id,
        entity_type=entity.kind,
        year_id=year_id,
        data_type=data_type.type,
    )
    data = data[data["measure_id"] == MEASURES["Incidence rate"]]
    return data


def extract_prevalence(
    entity: Cause | Sequela,
    location_id: list[int],
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:

    data = gbd.get_incidence_prevalence(
        entity_id=entity.gbd_id,
        location_id=location_id,
        entity_type=entity.kind,
        year_id=year_id,
        data_type=data_type.type,
    )
    data = data[data["measure_id"] == MEASURES["Prevalence"]]
    return data


def extract_birth_prevalence(
    entity: Cause | Sequela,
    location_id: list[int],
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_birth_prevalence(
        entity_id=entity.gbd_id,
        location_id=location_id,
        entity_type=entity.kind,
        year_id=year_id,
        data_type=data_type.type,
    )
    data = data[data["measure_id"] == MEASURES["Incidence rate"]]
    return data


def extract_disability_weight(
    entity: Sequela,
    location_id: list[int],
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    disability_weights = gbd.get_auxiliary_data(
        "disability_weight",
        entity.kind,
        "all",
        location_id,
    )
    data = disability_weights.loc[
        disability_weights.healthstate_id == entity.healthstate.gbd_id, :
    ]
    if year_id:  # if not pulling all years
        if isinstance(year_id, list):
            # Create a DataFrame from your year_ids
            year_df = pd.DataFrame({"year_id": year_id})
            # Perform a cross join by creating a temporary key to merge on
            data["key"] = 1
            year_df["key"] = 1
            # Merge to get the Cartesian product, then drop the key column
            data = data.drop("year_id", axis=1) if "year_id" in data.columns else data
            data = pd.merge(data, year_df, on="key").drop("key", axis=1)
        else:
            data["year_id"] = year_id
    return data


def extract_remission_rate(
    entity: Cause,
    location_id: list[int],
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_modelable_entity_draws(entity.me_id, location_id, year_id, data_type.type)
    data = data[data["measure_id"] == MEASURES["Remission rate"]]
    return data


def extract_deaths(
    entity: Cause,
    location_id: list[int],
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_codcorrect_draws(
        entity.gbd_id, location_id, year_id=year_id, data_type=data_type.type
    )
    data = data[data["measure_id"] == MEASURES["Deaths"]]
    return data


def extract_exposure(
    entity: RiskFactor | AlternativeRiskFactor,
    location_id: int,
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    if entity.kind == "risk_factor":
        data = gbd.get_exposure(entity.gbd_id, location_id, year_id, data_type.type)
        if entity.gbd_id == 341:
            data = process_kidney_dysfunction_exposure(data)
        allowable_measures = [
            MEASURES["Proportion"],
            MEASURES["Continuous"],
            MEASURES["Prevalence"],
        ]
        proper_measure_id = set(data["measure_id"]).intersection(allowable_measures)
        if len(proper_measure_id) != 1:
            raise DataAbnormalError(
                f"Exposure data have {len(proper_measure_id)} measure id(s). Data should have"
                f"exactly one id out of {allowable_measures} but came back with {proper_measure_id}."
            )
        else:
            data = data[data["measure_id"] == proper_measure_id.pop()]

    else:  # alternative_risk_factor
        data = gbd.get_auxiliary_data("exposure", entity.kind, entity.name, location_id)

    return data


def extract_exposure_standard_deviation(
    entity: RiskFactor | AlternativeRiskFactor,
    location_id: int,
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    if entity.kind == "risk_factor" and entity.name in OTHER_MEID:
        data = gbd.get_modelable_entity_draws(
            OTHER_MEID[entity.name], location_id, year_id, data_type.type
        )
    elif entity.kind == "risk_factor":
        data = gbd.get_exposure_standard_deviation(
            entity.gbd_id, location_id, year_id, data_type.type
        )
    else:  # alternative_risk_factor
        data = gbd.get_auxiliary_data(
            "exposure_standard_deviation", entity.kind, entity.name, location_id
        )
    return data


def extract_exposure_distribution_weights(
    entity: RiskFactor | AlternativeRiskFactor,
    location_id: int,
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_auxiliary_data(
        "exposure_distribution_weights", entity.kind, entity.name, location_id
    )
    return data


def extract_relative_risk(
    entity: RiskFactor,
    location_id: int,
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_relative_risk(entity.gbd_id, location_id, year_id, data_type.type)
    if not data["exposure"].isna().any() and data["parameter"].isna().all():
        data["parameter"] = data["exposure"]
        data["exposure"] = np.nan
    data = filter_to_most_detailed_causes(data)
    if entity.gbd_id == 136:  # non-exclusive breastfeeding
        data = data.loc[data["age_group_id"].isin([3, 388])]
    elif entity.gbd_id == 137:  # discontinued breastfeeding
        data = data.loc[data["age_group_id"].isin([238, 389])]
    return data


def extract_population_attributable_fraction(
    entity: RiskFactor | Etiology,
    location_id: int,
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_paf(entity.gbd_id, location_id, year_id, data_type.type)
    data = data[data["metric_id"] == METRICS["Percent"]]
    data = data[data["measure_id"].isin([MEASURES["YLDs"], MEASURES["YLLs"]])]
    data = filter_to_most_detailed_causes(data)
    # clip PAFs between 0 and 1 (data outside these bounds is expected from GBD)
    draw_cols = [col for col in data.columns if col.startswith("draw_")]
    data.loc[:, draw_cols] = data[draw_cols].clip(lower=0)
    data.loc[:, draw_cols] = data[draw_cols].clip(upper=1)
    return data


def extract_mediation_factors(
    entity: RiskFactor,
    location_id: int,
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_auxiliary_data("mediation_factor", entity.kind, entity.name, location_id)
    return data


def extract_estimate(
    entity: Covariate,
    location_id: int,
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_covariate_estimate(int(entity.gbd_id), location_id, year_id)
    return data


# FIXME: can this be deleted? 'utilization_rate' is not in the get_data() mapping.
def extract_utilization_rate(
    entity: HealthcareEntity,
    location_id: int,
    year_id: Optional[Union[int, str, List[int]]],
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_modelable_entity_draws(
        entity.gbd_id, location_id, year_id=None, data_type=data_type.type
    )
    return data


def extract_structure(
    entity: Population,
    location_id: int,
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_population(location_id, year_id=year_id)
    return data


def extract_theoretical_minimum_risk_life_expectancy(
    entity: Population,
    location_id: int,
    year_id: int | str | list[int] | None,
    data_type: DataType,
) -> pd.DataFrame:
    data = gbd.get_theoretical_minimum_risk_life_expectancy()
    return data



from gbd_mapping.base_template import GbdRecord, ModelableEntity


class HealthTechnology(ModelableEntity):
    """Container for health technology ids and data."""

    __slots__ = ("name", "kind", "gbd_id")


class HealthTechnologies(GbdRecord):
    """Holder of healthcare technologies."""

    __slots__ = ("hypertension_drugs",)

    def __init__(self, hypertension_drugs: HealthTechnology):
        self.hypertension_drugs = hypertension_drugs



import numpy as np
import pandas as pd
import pytest

from tests.conftest import NO_GBD_ACCESS
from vivarium_inputs.globals import SEXES, DataAbnormalError, DataDoesNotExistError
from vivarium_inputs.validation import raw

LOCATION_ID = 100
ESTIMATION_YEARS = tuple(list(range(1990, 2015, 5)) + list(range(2019, 2022)))
AGE_GROUP_IDS = tuple([2, 3, 388, 389, 238, 34])
PARENT_LOCATIONS = (1, 5, 10)


@pytest.fixture
def mock_validation_context():
    context = raw.RawValidationContext(
        location_id=[LOCATION_ID],
        estimation_years=list(ESTIMATION_YEARS),
        age_group_ids=list(AGE_GROUP_IDS),
        sexes=SEXES,
        parent_locations={LOCATION_ID: list(PARENT_LOCATIONS)},
    )
    return context


@pytest.fixture
def measures_mock(mocker):
    return mocker.patch("vivarium_inputs.validation.raw.MEASURES", {"A": 1, "B": 2})


@pytest.fixture
def metrics_mock(mocker):
    return mocker.patch("vivarium_inputs.validation.raw.METRICS", {"A": 1, "B": 2})


@pytest.mark.parametrize(
    "mort, morb, yld_only, yll_only, match",
    [
        ([0, 1, 1], [0, 1, 1], True, False, "set to 0"),
        ([1, 1, 0], [1, 1, 1], True, False, "only one"),
        ([1, 1, 1], [0, 0, 0], True, False, "restricted to yld_only"),
        ([0, 0, 0], [1, 1, 1], False, True, "restricted to yll_only"),
        ([1, 0, 0], [0, 1, 1], True, False, "is restricted to yld_only"),
        ([1, 0, 2], [0, 1, 1], False, False, "outside the expected"),
    ],
)
def test_check_mort_morb_flags_fail(mort, morb, yld_only, yll_only, match):
    data = pd.DataFrame({"mortality": mort, "morbidity": morb})
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_mort_morb_flags(data, yld_only, yll_only)


@pytest.mark.parametrize(
    "mort, morb, yld_only, yll_only",
    [
        ([1, 1, 1], [1, 1, 1], True, False),
        ([1, 1, 1], [1, 1, 1], False, False),
        (
            [0, 0, 0],
            [1, 1, 1],
            True,
            False,
        ),
        (
            [1, 1, 1],
            [0, 0, 0],
            False,
            True,
        ),
        ([1, 0, 1], [0, 1, 0], False, False),
    ],
)
def test_check_mort_morb_flags_pass(mort, morb, yld_only, yll_only):
    data = pd.DataFrame({"mortality": mort, "morbidity": morb})
    raw.check_mort_morb_flags(data, yld_only, yll_only)


@pytest.mark.parametrize(
    "years, bin_type",
    [
        (range(1950, 2023), "annual"),
        (range(1990, 2022), "annual"),
        (list(range(1990, 2015, 5)) + list(range(2019, 2022)), "binned"),
    ],
    ids=["annual with extra", "annual", "binned"],
)
def test_check_years_pass(mock_validation_context, years, bin_type):
    df = pd.DataFrame({"year_id": years})
    raw.check_years(df, mock_validation_context, bin_type)


@pytest.mark.parametrize(
    "years, bin_type, match",
    [
        ([1990, 1992], "annual", "missing"),
        ([1990, 1995], "binned", "missing"),
        (list(range(1990, 2015, 5)) + list(range(2019, 2023)), "binned", "extra"),
    ],
    ids=["annual with gap", "binned with gap", "binned with extra"],
)
def test_check_years_fail(mock_validation_context, years, bin_type, match):
    df = pd.DataFrame({"year_id": years})
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_years(df, mock_validation_context, bin_type)


@pytest.mark.parametrize("location_id", list(range(2, 10)))
def test_check_location_pass(mock_validation_context, location_id):
    mock_validation_context["location_id"] = [location_id]
    df = pd.DataFrame({"location_id": [location_id] * 5})
    raw.check_location(df, mock_validation_context)


@pytest.mark.parametrize("location_id", PARENT_LOCATIONS)
def test_check_location_parent_pass(mock_validation_context, location_id):
    df = pd.DataFrame({"location_id": [location_id]})
    raw.check_location(df, mock_validation_context)


@pytest.mark.parametrize("location_id", list(range(20, 25)))
def test_check_location_fail(mock_validation_context, location_id):
    df = pd.DataFrame({"location_id": [location_id]})
    with pytest.raises(DataAbnormalError, match="Data pulled for "):
        raw.check_location(df, mock_validation_context)


@pytest.mark.parametrize("columns", [["a", "b"], ["c", "d", "e"], ["a"], []])
def test_check_columns_pass(columns):
    raw.check_columns(columns, columns)


@pytest.mark.parametrize("columns", [["a", "b"], ["c", "d", "e"], ["a"], []])
def test_check_columns_extra_warn(columns, caplog, match="Data returned extra columns"):
    raw.check_columns(columns, columns + ["extra"])
    assert match in caplog.text


@pytest.mark.parametrize("columns", [["a", "b"], ["c", "d", "e"], ["a"]])
def test_check_columns_missing_fail(columns):
    with pytest.raises(DataAbnormalError, match="missing columns"):
        raw.check_columns(columns, columns[:-1])


@pytest.mark.parametrize(
    "data",
    [
        pd.DataFrame({"a": [1], "b": [np.nan], "c": [0]}),
        pd.DataFrame({"a": [np.nan], "b": [np.nan]}),
        pd.DataFrame({"a": [np.inf], "b": [1], "c": [0]}),
        pd.DataFrame({"a": [np.inf], "b": [np.inf]}),
        pd.DataFrame(columns=["a", "b"]),
    ],
    ids=["Single NaN", "All NaN", "Single Inf", "All Inf", "Empty"],
)
def test_check_data_exist_always_fail(data):
    with pytest.raises(DataDoesNotExistError):
        raw.check_data_exist(data, value_columns=data.columns, zeros_missing=False)

    assert (
        raw.check_data_exist(
            data, value_columns=data.columns, zeros_missing=False, error=False
        )
        is False
    )

    with pytest.raises(DataDoesNotExistError):
        raw.check_data_exist(data, value_columns=data.columns, zeros_missing=True)

    assert (
        raw.check_data_exist(
            data, value_columns=data.columns, zeros_missing=True, error=False
        )
        is False
    )


def test_check_data_exist_fails_only_on_missing_zeros():
    df = pd.DataFrame({"a": [0, 0], "b": [0, 0]})

    raw.check_data_exist(df, value_columns=df.columns, zeros_missing=False)

    with pytest.raises(DataDoesNotExistError):
        raw.check_data_exist(df, value_columns=df.columns, zeros_missing=True)


def test_check_data_exist_value_columns():
    df = pd.DataFrame({"a": [0, 0], "b": [0, 0], "c": [1, 1]})

    assert raw.check_data_exist(
        df, value_columns=["a", "b", "c"], zeros_missing=True, error=False
    )

    assert (
        raw.check_data_exist(df, value_columns=["a", "b"], zeros_missing=True, error=False)
        is False
    )


@pytest.mark.parametrize(
    "data",
    [
        pd.DataFrame({"a": [1], "b": [1], "c": [0]}),
        pd.DataFrame({"a": [0], "b": [1]}),
        pd.DataFrame({"a": [0.0001], "b": [0], "c": [0]}),
        pd.DataFrame({"a": [1000], "b": [-1000]}),
    ],
)
def test_check_data_exist_pass(data):
    assert raw.check_data_exist(
        data, zeros_missing=True, value_columns=data.columns, error=False
    )


@pytest.mark.parametrize(
    "test_age_ids, start, end, match",
    [
        ([2, 3, 388, 100], None, None, "invalid"),
        ([2, 3, 388, 238], None, None, "non-contiguous"),
        ([2, 3, 388, 100], 2, 388, "invalid"),
        ([2, 3, 388, 238], 2, 388, "non-contiguous"),
    ],
)
@pytest.mark.skipif(
    NO_GBD_ACCESS, reason="Cannot run these tests without vivarium_gbd_access"
)
def test_check_age_group_ids_fail(mock_validation_context, test_age_ids, start, end, match):
    df = pd.DataFrame({"age_group_id": test_age_ids})
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_age_group_ids(df, mock_validation_context, start, end)


@pytest.mark.parametrize(
    "test_age_ids, start, end, match",
    [
        ([2, 3], 2, 388, "contain all age groups in restriction range"),
        ([2, 3, 388], 2, 3, "additional age groups"),
        ([2, 3, 388, 389, 238], 2, 388, "additional age groups"),
        ([2, 3, 388], 2, 238, "contain all age groups in restriction range"),
    ],
)
@pytest.mark.skipif(
    NO_GBD_ACCESS, reason="Cannot run these tests without vivarium_gbd_access"
)
def test_check_age_group_ids_warn(
    mock_validation_context, caplog, test_age_ids, start, end, match
):
    df = pd.DataFrame({"age_group_id": test_age_ids})
    raw.check_age_group_ids(df, mock_validation_context, start, end)
    assert match in caplog.text


@pytest.mark.parametrize(
    "test_age_ids, start, end",
    [([2, 3, 388], 2, 388), ([2, 3, 388], None, None), ([2, 3, 388, 389, 238], 2, 238)],
)
@pytest.mark.skipif(
    NO_GBD_ACCESS, reason="Cannot run these tests without vivarium_gbd_access"
)
def test_check_age_group_ids_pass(mock_validation_context, test_age_ids, start, end, recwarn):
    df = pd.DataFrame({"age_group_id": test_age_ids})
    raw.check_age_group_ids(df, mock_validation_context, start, end)

    assert len(recwarn) == 0, "An unexpected warning was raised."


@pytest.mark.parametrize(
    "sex_ids, male, female, both",
    [
        ([1, 2, 12], True, False, False),
        ([1, 2, 3, 3, 0], True, True, True),
        ([5], False, False, False),
    ],
)
def test_check_sex_ids_fail(mock_validation_context, sex_ids, male, female, both):
    df = pd.DataFrame({"sex_id": sex_ids})
    with pytest.raises(DataAbnormalError):
        raw.check_sex_ids(df, mock_validation_context, male, female, both)


@pytest.mark.parametrize(
    "sex_ids, male, female, both, warn_extra, warn_missing",
    [
        ([1, 2, 1], True, True, True, False, True),
        ([1, 2, 3, 2], True, True, False, True, False),
        ([1], False, True, False, True, True),
    ],
)
def test_check_sex_ids_warn(
    mock_validation_context, caplog, sex_ids, male, female, both, warn_extra, warn_missing
):
    df = pd.DataFrame({"sex_id": sex_ids})
    raw.check_sex_ids(df, mock_validation_context, male, female, both)
    assert warn_extra == (
        "extra" in caplog.text
    ), "The expected warning message was not raised for extra sex ids"
    assert warn_missing == (
        "missing" in caplog.text
    ), "The expected warning message was not raised for missing sex ids"


@pytest.mark.parametrize(
    "sex_ids, male, female, both",
    [
        ([1, 1, 1], True, False, False),
        ([1, 2, 2, 2], True, True, False),
        ([1, 2, 3, 3, 2, 1], True, True, True),
        ([], False, False, False),
    ],
)
def test_check_sex_ids_pass(mock_validation_context, sex_ids, male, female, both, recwarn):
    df = pd.DataFrame({"sex_id": sex_ids})
    raw.check_sex_ids(df, mock_validation_context, male, female, both)

    assert len(recwarn) == 0, "An unexpected warning was raised."


test_data = [
    (
        pd.DataFrame({"age_group_id": [2, 3, 388], "a": 1, "b": 0}),
        2,
        389,
        ["a", "b"],
        "missing",
    ),
    (
        pd.DataFrame({"age_group_id": [2, 3], "a": [0, 0], "b": [0, 0]}),
        2,
        3,
        ["a", "b"],
        "missing for all age groups",
    ),
]


@pytest.mark.parametrize("data, start, end, val_cols, match", test_data)
def test_check_age_restrictions_fail(
    mock_validation_context, data, start, end, val_cols, match
):
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_age_restrictions(
            data, mock_validation_context, start, end, value_columns=val_cols
        )


test_data = [
    (
        pd.DataFrame({"age_group_id": [2, 3, 388, 389, 238], "a": 1, "b": 0}),
        2,
        389,
        ["a", "b"],
    ),
    (pd.DataFrame({"age_group_id": [2, 3, 4], "a": [1, 1, 1], "b": [2, 3, 0]}), 2, 3, ["a"]),
]


@pytest.mark.parametrize("data, start, end, val_cols", test_data)
def test_check_age_restrictions_warn(
    mock_validation_context, caplog, data, start, end, val_cols
):
    raw.check_age_restrictions(
        data, mock_validation_context, start, end, value_columns=val_cols
    )
    assert "also included" in caplog.text


test_data = [
    (pd.DataFrame({"age_group_id": [2, 3, 388], "a": 1, "b": 0}), 2, 388, ["a", "b"]),
    (pd.DataFrame({"age_group_id": [2, 3], "a": [1, 0], "b": [1, 0.1]}), 2, 3, ["a", "b"]),
    (
        pd.DataFrame({"age_group_id": [2, 3, 388], "a": [1, 1, 0], "b": [2, 3, 0]}),
        2,
        388,
        ["a"],
    ),
]


@pytest.mark.parametrize("data, start, end, val_cols", test_data)
def test_check_age_restrictions_pass(mock_validation_context, data, start, end, val_cols):
    raw.check_age_restrictions(
        data, mock_validation_context, start, end, value_columns=val_cols
    )


test_data = [
    (
        pd.DataFrame({"sex_id": [1, 1], "a": 0, "b": 0, "c": 1}),
        True,
        False,
        ["a", "b"],
        "missing data values for males",
    ),
    (
        pd.DataFrame({"sex_id": [2, 2, 2, 3], "a": 0, "b": 0}),
        False,
        True,
        ["a", "b"],
        "missing data values for females",
    ),
    (
        pd.DataFrame({"sex_id": [3], "a": [0], "b": [0]}),
        False,
        False,
        ["a", "b"],
        "values for both",
    ),
    (
        pd.DataFrame({"sex_id": [1, 2], "a": [0, 1], "b": [0, 1]}),
        False,
        False,
        ["a", "b"],
        "values for both",
    ),
]


@pytest.mark.parametrize("data, male, female, val_cols, match", test_data)
def test_check_sex_restrictions_fail(
    mock_validation_context, data, male, female, val_cols, match
):
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_sex_restrictions(
            data, mock_validation_context, male, female, value_columns=val_cols
        )


test_data = [
    (
        pd.DataFrame({"sex_id": [1, 2], "a": 1, "b": 0}),
        True,
        False,
        ["a", "b"],
        "non-male sex ids",
    ),
    (
        pd.DataFrame({"sex_id": [1, 2], "a": 1, "b": 0}),
        False,
        True,
        ["a", "b"],
        "non-female sex ids",
    ),
]


@pytest.mark.parametrize("data, male, female, val_cols, match", test_data)
def test_check_sex_restrictions_warn(
    mock_validation_context, caplog, data, male, female, val_cols, match
):
    raw.check_sex_restrictions(
        data, mock_validation_context, male, female, value_columns=val_cols
    )
    assert match in caplog.text


test_data = [
    (pd.DataFrame({"sex_id": [1, 1], "a": 1, "b": 0, "c": 0}), True, False, ["a", "b"]),
    (pd.DataFrame({"sex_id": [2, 2], "a": 1, "b": 0}), False, True, ["a", "b"]),
    (pd.DataFrame({"sex_id": [3, 3, 3], "a": 1, "b": 0}), False, False, ["a"]),
    (pd.DataFrame({"sex_id": [1, 2], "a": [1, 0], "b": [0, 1]}), False, False, ["a", "b"]),
    (pd.DataFrame({"sex_id": [3], "a": [0], "b": [1]}), False, False, ["a", "b"]),
]


@pytest.mark.parametrize("data, male, female, val_cols", test_data)
def test_check_sex_restrictions_pass(mock_validation_context, data, male, female, val_cols):
    raw.check_sex_restrictions(
        data, mock_validation_context, male, female, value_columns=val_cols
    )


@pytest.mark.parametrize(
    "m_ids, expected, match",
    [([1, 1, 1, 12], ["A"], "multiple"), ([3, 3, 3], ["A", "B"], "not in the expected")],
)
def test_check_measure_id_fail(m_ids, expected, match, measures_mock):
    df = pd.DataFrame({"measure_id": m_ids})
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_measure_id(df, expected)


@pytest.mark.parametrize("m_ids, expected", [([1], ["A"]), ([2, 2, 2, 2], ["A", "B"])])
def test_check_measure_id_pass(m_ids, expected, measures_mock):
    df = pd.DataFrame({"measure_id": m_ids})
    raw.check_measure_id(df, expected)


@pytest.mark.parametrize("m_ids, expected", [([1, 1, 1, 12], "A"), ([3, 3, 3], "A")])
def test_check_metric_id_fail(m_ids, expected, metrics_mock):
    df = pd.DataFrame({"metric_id": m_ids})
    with pytest.raises(DataAbnormalError):
        raw.check_metric_id(df, expected)


@pytest.mark.parametrize("m_ids, expected", [([1], "A"), ([2, 2, 2, 2], "B")])
def test_check_metric_id_pass(m_ids, expected, metrics_mock):
    df = pd.DataFrame({"metric_id": m_ids})
    raw.check_metric_id(df, expected)



import pytest
from gbd_mapping import ModelableEntity, causes, covariates, risk_factors
from pytest_mock import MockerFixture

from tests.conftest import is_not_implemented
from tests.mocked_gbd import (
    MOST_RECENT_YEAR,
    get_mocked_location_ids,
    mock_vivarium_gbd_access,
)
from vivarium_inputs import core
from vivarium_inputs.globals import Population
from vivarium_inputs.utilities import DataType, DataTypeNotImplementedError

LOCATION = "India"

CAUSES = [
    (
        causes.measles,
        [
            "incidence_rate",
            "raw_incidence_rate",
            "prevalence",
            "disability_weight",
            "cause_specific_mortality_rate",
            "excess_mortality_rate",
            "deaths",
        ],
    ),
    (
        causes.diarrheal_diseases,
        [
            "incidence_rate",
            "raw_incidence_rate",
            "prevalence",
            "disability_weight",
            "remission_rate",
            "cause_specific_mortality_rate",
            "excess_mortality_rate",
            "deaths",
        ],
    ),
    (
        causes.diabetes_mellitus_type_2,
        [
            "incidence_rate",
            "raw_incidence_rate",
            "prevalence",
            "disability_weight",
            "cause_specific_mortality_rate",
            "excess_mortality_rate",
            "deaths",
        ],
    ),
]
CAUSE_MEASURES = [
    "incidence_rate",
    "raw_incidence_rate",
    "prevalence",
    "birth_prevalence",
    "disability_weight",
    "remission_rate",
    "cause_specific_mortality_rate",
    "excess_mortality_rate",
    "deaths",
]


@pytest.mark.parametrize("entity_details", CAUSES, ids=lambda x: x[0].name)
@pytest.mark.parametrize("measure", CAUSE_MEASURES, ids=lambda x: x)
@pytest.mark.parametrize(
    "years", [None, 2019, 1900, [2019], [2019, 2020], "all"], ids=lambda x: str(x)
)
@pytest.mark.parametrize(
    "data_type", ["draws", "means", ["draws", "means"]], ids=lambda x: str(x)
)
def test_year_id_causelike(
    entity_details: tuple[ModelableEntity, list[str]],
    measure: str,
    years: int | list[int] | str | None,
    data_type: str | list[str],
    mocker: MockerFixture,
):
    if years == "all" and measure == "remission_rate":
        # remission rates and relative risks have binned years and for
        # "all" years will use central comp's `core-maths` library which, like
        # vivarium_gbd_access, is hosted on bitbucket and so cannot be accessed
        # from github-actions.
        pytest.skip("Expected to fail - see test_xfailed test")

    entity, entity_expected_measures = entity_details
    if measure in entity_expected_measures:
        check_year_in_data(entity, measure, LOCATION, years, data_type, mocker)


RISKS = [
    (
        risk_factors.high_systolic_blood_pressure,
        [
            "exposure",
            "exposure_standard_deviation",
            "exposure_distribution_weights",
            "relative_risk",
            "population_attributable_fraction",
        ],
    ),
    (
        risk_factors.low_birth_weight_and_short_gestation,
        [
            "exposure",
            "relative_risk",
            "population_attributable_fraction",
        ],
    ),
]
RISK_MEASURES = [
    "exposure",
    "exposure_standard_deviation",
    "exposure_distribution_weights",
    "relative_risk",
    "population_attributable_fraction",
]


@pytest.mark.parametrize("entity_details", RISKS, ids=lambda x: x[0].name)
@pytest.mark.parametrize("measure", RISK_MEASURES, ids=lambda x: x)
@pytest.mark.parametrize(
    "years", [None, 2019, 1900, [2019], [2019, 2020], "all"], ids=lambda x: str(x)
)
@pytest.mark.parametrize(
    "data_type", ["draws", "means", ["draws", "means"]], ids=lambda x: str(x)
)
def test_year_id_risklike(
    entity_details: tuple[ModelableEntity, list[str]],
    measure: str,
    years: int | list[int] | str | None,
    data_type: str | list[str],
    mocker: MockerFixture,
):
    if years == "all" and measure == "relative_risk":
        # remission rates and relative risks have binned years and for
        # "all" years will use central comp's `core-maths` library which, like
        # vivarium_gbd_access, is hosted on bitbucket and so cannot be accessed
        # from github-actions.
        pytest.skip("Expected to fail - see test_xfailed test")
    if (
        measure == "relative_risk"
        and entity_details[0].name == "high_systolic_blood_pressure"
        and data_type == "draws"
    ):
        pytest.skip("FIXME: [mic-5542] continuous rrs cannot validate")
    entity, entity_expected_measures = entity_details
    if measure in entity_expected_measures:
        check_year_in_data(entity, measure, LOCATION, years, data_type, mocker)


@pytest.mark.xfail(raises=ModuleNotFoundError, reason="Cannot import core-maths", strict=True)
@pytest.mark.parametrize(
    "entity, measure",
    [
        [causes.diarrheal_diseases, "remission_rate"],
        [risk_factors.high_systolic_blood_pressure, "relative_risk"],
        [risk_factors.low_birth_weight_and_short_gestation, "relative_risk"],
    ],
)
def test_xfailed(
    entity: ModelableEntity,
    measure: str,
    mocker: MockerFixture,
):
    """We expect failures when trying to interpolate 'all' years for binned measures

    Notes
    -----
    These test parameterizations are a subset of others in this test module
    that are simply marked as 'skip'.
    """
    if measure == "relative_risk" and entity.name == "high_systolic_blood_pressure":
        pytest.skip("FIXME: [mic-5542] continuous rrs cannot validate")
    check_year_in_data(entity, measure, LOCATION, "all", "draws", mocker)


COVARIATES = [
    covariates.systolic_blood_pressure_mmhg,
]
COVARIATE_MEASURES = ["estimate"]


@pytest.mark.parametrize("entity", COVARIATES, ids=lambda x: x.name)
@pytest.mark.parametrize("measure", COVARIATE_MEASURES, ids=lambda x: x)
@pytest.mark.parametrize(
    "years", [None, 2019, 1900, [2019], [2019, 2020], "all"], ids=lambda x: str(x)
)
@pytest.mark.parametrize(
    "data_type", ["draws", "means", ["draws", "means"]], ids=lambda x: str(x)
)
def test_year_id_covariatelike(
    entity: ModelableEntity,
    measure: str,
    years: int | list[int] | str | None,
    data_type: str | list[str],
    mocker: MockerFixture,
):
    check_year_in_data(entity, measure, LOCATION, years, data_type, mocker)


@pytest.mark.parametrize("measure", ["structure", "demographic_dimensions"], ids=lambda x: x)
@pytest.mark.parametrize(
    "years", [None, 2019, 1900, [2019], [2019, 2020], "all"], ids=lambda x: str(x)
)
@pytest.mark.parametrize(
    "data_type", ["draws", "means", ["draws", "means"]], ids=lambda x: str(x)
)
def test_year_id_population(
    measure: str,
    years: int | list[int] | str | None,
    data_type: str | list[str],
    mocker: MockerFixture,
):
    pop = Population()
    check_year_in_data(pop, measure, LOCATION, years, data_type, mocker)


@pytest.mark.parametrize("entity_details", CAUSES, ids=lambda x: x[0].name)
@pytest.mark.parametrize("measure", CAUSE_MEASURES, ids=lambda x: x)
@pytest.mark.parametrize(
    "locations",
    [
        "Ethiopia",  # 179
        179,
        ["Ethiopia", "Nigeria"],  # [179, 214]
        [179, 214],
        [179, "Nigeria"],
    ],
    ids=lambda x: str(x),
)
@pytest.mark.parametrize(
    "data_type", ["draws", "means", ["draws", "means"]], ids=lambda x: str(x)
)
def test_multiple_locations_causelike(
    entity_details: tuple[ModelableEntity, list[str]],
    measure: str,
    locations: str | int | list[str | int],
    data_type: str | list[str],
    mocker: MockerFixture,
):
    year = MOST_RECENT_YEAR
    location_id_mapping = {
        "Ethiopia": 179,
        "Nigeria": 214,
    }
    entity, entity_expected_measures = entity_details
    if is_not_implemented(data_type, measure):
        with pytest.raises(DataTypeNotImplementedError):
            data_type = DataType(measure, data_type)
            mocker.patch(
                "vivarium_inputs.utility_data.get_raw_location_ids",
                return_value=get_mocked_location_ids(),
            )
            core.get_data(entity, measure, locations, year, data_type)
    else:
        if measure not in entity_expected_measures:
            with pytest.raises(Exception):
                data_type = DataType(measure, data_type)
                core.get_data(entity, measure, locations, year, data_type)
        else:
            mocked_funcs = mock_vivarium_gbd_access(
                entity, measure, locations, year, data_type, mocker
            )
            data_type = DataType(measure, data_type)
            df = core.get_data(entity, measure, locations, year, data_type)
            for mocked_func in mocked_funcs:
                assert mocked_func.called_once()
            if not isinstance(locations, list):
                locations = [locations]
            location_ids = {
                (
                    location_id_mapping[item]
                    if isinstance(item, str) and item in location_id_mapping
                    else item
                )
                for item in locations
            }
            assert set(df.index.get_level_values("location_id")) == set(location_ids)


# TODO: Should we add the location tests for other entity types?


####################
# Helper functions #
####################


def check_year_in_data(
    entity: ModelableEntity,
    measure: str,
    location: str,
    years: int | list[int] | str | None,
    data_type: str | list[str],
    mocker: MockerFixture,
):
    if is_not_implemented(data_type, measure):
        with pytest.raises(DataTypeNotImplementedError):
            data_type = DataType(measure, data_type)
            mocker.patch(
                "vivarium_inputs.utility_data.get_raw_location_ids",
                return_value=get_mocked_location_ids(),
            )
            core.get_data(entity, measure, location, years, data_type)
    else:
        mocked_funcs = mock_vivarium_gbd_access(
            entity, measure, location, years, data_type, mocker
        )
        data_type = DataType(measure, data_type)
        if isinstance(years, list):
            df = core.get_data(entity, measure, location, years, data_type)
            assert set(df.index.get_level_values("year_id")) == set(years)
            for mocked_func in mocked_funcs:
                assert mocked_func.called_once()
        # years expected to be 1900, 2019, None, or "all"
        elif years != 1900:
            df = core.get_data(entity, measure, location, years, data_type)
            if years == None:
                assert set(df.index.get_level_values("year_id")) == set([MOST_RECENT_YEAR])
            elif years == "all":
                assert set(df.index.get_level_values("year_id")) == set(range(1990, 2023))
            else:  # a single (non-1900) year
                assert set(df.index.get_level_values("year_id")) == set([years])
            for mocked_func in mocked_funcs:
                assert mocked_func.called_once()
        else:
            with pytest.raises(ValueError, match="years must be in"):
                core.get_data(entity, measure, location, years, data_type)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
from pathlib import Path

from sphinx.ext.intersphinx import missing_reference

import vivarium_inputs

base_dir = Path(vivarium_inputs.__file__).parent

about = {}
with (base_dir / "__about__.py").open() as f:
    exec(f.read(), about)

sys.path.insert(0, str(Path("..").resolve()))

# -- Project information -----------------------------------------------------

project = about["__title__"]
copyright = f'2023, {about["__author__"]}'
author = about["__author__"]

# The short X.Y version.
version = vivarium_inputs.__version__
# The full version, including alpha/beta/rc tags.
release = vivarium_inputs.__version__


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.

needs_sphinx = "4.0"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme_path = ["_theme"]
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "globaltoc.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = f'{about["__title__"]}doc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        f'{about["__title__"]}.tex',
        f'{about["__title__"]} Documentation',
        about["__author__"],
        "manual",
    ),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, f'{about["__title__"]}', f'{about["__title__"]} Documentation', [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        f'{about["__title__"]}',
        f'{about["__title__"]} Documentation',
        author,
        f'{about["__title__"]}',
        about["__summary__"],
        "Miscellaneous",
    ),
]

# Other docs we can link to
intersphinx_mapping = {
    "python": ("https://docs.python.org/3.11", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "tables": ("https://www.pytables.org/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "networkx": ("https://networkx.org/documentation/stable/", None),
    "vivarium": ("https://vivarium.readthedocs.io/en/latest/", None),
    "gbd_mapping": ("https://vivarium.readthedocs.io/projects/gbd-mapping/en/latest/", None),
}


# -- Autodoc configuration ------------------------------------------------

autodoc_default_options = {
    # Automatically document members (e.g. classes in a module,
    # methods in a class, etc.)
    "members": True,
    # Order of items documented is determined by the order
    # of appearance in the source code
    "member-order": "bysource",
    # Generate docs even if an item has no docstring.
    "undoc-members": True,
    # Don't document things with a leading underscore.
    "private-members": False,
}
# Display type hints in the description instead of the signature.
autodoc_typehints = "description"


# -- nitpicky mode --------------------------------------------------------
# Ensures that all references in the docs resolve.

nitpicky = True

nitpick_ignore = []
for line in open("../nitpick-exceptions"):
    if line.strip() == "" or line.startswith("#"):
        continue
    dtype, target = line.split(None, 1)
    target = target.strip()
    nitpick_ignore.append((dtype, target))



"""End-to-end tests for interface.get_measure()."""

from __future__ import annotations

import datetime
from functools import partial

import pandas as pd
import pytest
from gbd_mapping import (
    Cause,
    Covariate,
    RiskFactor,
    Sequela,
    causes,
    covariates,
    risk_factors,
    sequelae,
)
from layered_config_tree import LayeredConfigTree
from pytest_mock import MockerFixture

from tests.conftest import NO_GBD_ACCESS, is_not_implemented
from tests.mocked_gbd import (
    MOST_RECENT_YEAR,
    get_mocked_age_bins,
    mock_vivarium_gbd_access,
)
from vivarium_inputs import utility_data
from vivarium_inputs.globals import NON_STANDARD_MEASURES, SUPPORTED_DATA_TYPES
from vivarium_inputs.interface import get_measure
from vivarium_inputs.utilities import DataTypeNotImplementedError

SLOW_TEST_DAY = "Sunday"  # Day to run very slow tests, e.g. PAFs and RRs


LOCATION = "India"

# TODO [MIC-5448]: Move to vivarium_testing_utilties
@pytest.fixture(autouse=True)
def no_cache(mocker: MockerFixture) -> None:
    """Mock out the cache so that we always pull data."""

    if not NO_GBD_ACCESS:
        mocker.patch(
            "vivarium_gbd_access.utilities.get_input_config",
            return_value=LayeredConfigTree({"input_data": {"cache_data": False}}),
        )


CAUSES = [
    # (entity, applicable_measures)
    # NOTE: 'raw_incidence_rate' and 'deaths' should not be called directly from `get_measure()`
    # because there are no implemented validations for them.
    (
        causes.measles,
        [
            "incidence_rate",
            "prevalence",
            "disability_weight",
            "cause_specific_mortality_rate",
            "excess_mortality_rate",
        ],
    ),
    (
        causes.diarrheal_diseases,
        [
            "incidence_rate",
            "prevalence",
            "disability_weight",
            "remission_rate",
            "cause_specific_mortality_rate",
            "excess_mortality_rate",
        ],
    ),
    (
        causes.diabetes_mellitus_type_2,
        [
            "incidence_rate",
            "prevalence",
            "disability_weight",
            "cause_specific_mortality_rate",
            "excess_mortality_rate",
        ],
    ),
]
CAUSE_MEASURES = [
    "incidence_rate",
    "prevalence",
    "birth_prevalence",
    "disability_weight",
    "remission_rate",
    "cause_specific_mortality_rate",
    "excess_mortality_rate",
]


@pytest.mark.parametrize("entity_details", CAUSES, ids=lambda x: x[0].name)
@pytest.mark.parametrize("measure", CAUSE_MEASURES, ids=lambda x: x)
@pytest.mark.parametrize(
    "data_type", ["means", "draws", ["means", "draws"]], ids=lambda x: str(x)
)
@pytest.mark.parametrize(
    "mock_gbd", [True, False], ids=lambda x: "mocked" if x else "unmocked"
)
def test_get_measure_causelike(
    entity_details: tuple[Cause, list[str]],
    measure: str,
    data_type: str | list[str],
    mock_gbd: bool,
    runslow: bool,
    mocker: MockerFixture,
):
    """Test get_measure().

    We are parametrizing over various things.
    - entity_details: A tuple of an entity and a list of measures that are applicable to that entity.
    - measure: The possible measures for that entity type (e.g. cause, sequela, etc)
    - data_type: The data type request (means or draws)
    - mock_gbd: Whether to mock calls to vivarium_gbd_access or not. We do this because
        getting the real data tends to take a long time and so we want to only run that
        when requesting a --runslow test. Note that we do not attempt to get more granular
        with slow runs; mocked tests will always run and unmocked tests will only run
        if --runslow is passed

    Notes
    -----
    The --runslow flag is automatically passed into these tests as an argument; we do
    not mark the tests themselves as slow (because we want more granularity, i.e.
    mocked tests are not slow but unmocked tests are).
    """
    # Test-specific fixme skips
    if measure == "birth_prevalence":
        pytest.skip("FIXME: need to find causes with birth prevalence")

    run_test(entity_details, measure, data_type, mock_gbd, runslow, mocker)


SEQUELAE = [
    (
        sequelae.hiv_aids_drug_susceptible_tuberculosis_without_anemia,
        [
            "incidence_rate",
            "prevalence",
            "birth_prevalence",
            "disability_weight",
        ],
    ),
]
SEQUELA_MEASURES = [
    "incidence_rate",
    "prevalence",
    "birth_prevalence",
    "disability_weight",
]


@pytest.mark.parametrize("entity_details", SEQUELAE, ids=lambda x: x[0].name)
@pytest.mark.parametrize("measure", SEQUELA_MEASURES, ids=lambda x: x)
@pytest.mark.parametrize(
    "data_type", ["means", "draws", ["means", "draws"]], ids=lambda x: str(x)
)
@pytest.mark.parametrize(
    "mock_gbd", [True, False], ids=lambda x: "mocked" if x else "unmocked"
)
def test_get_measure_sequelalike(
    entity_details: tuple[Sequela, list[str]],
    measure: str,
    data_type: str | list[str],
    mock_gbd: bool,
    runslow: bool,
    mocker: MockerFixture,
):
    """Test get_measure().

    We are parametrizing over various things.
    - entity_details: A tuple of an entity and a list of measures that are applicable to that entity.
    - measure: The possible measures for that entity type (e.g. cause, sequela, etc)
    - data_type: The data type request (means or draws)
    - mock_gbd: Whether to mock calls to vivarium_gbd_access or not. We do this because
        getting the real data tends to take a long time and so we want to only run that
        when requesting a --runslow test. Note that we do not attempt to get more granular
        with slow runs; mocked tests will always run and unmocked tests will only run
        if --runslow is passed

    Notes
    -----
    The --runslow flag is automatically passed into these tests as an argument; we do
    not mark the tests themselves as slow (because we want more granularity, i.e.
    mocked tests are not slow but unmocked tests are).
    """

    # Test-specific fixme skips
    if measure == "birth_prevalence":
        pytest.skip("FIXME: need to find sequelae with birth prevalence")

    run_test(entity_details, measure, data_type, mock_gbd, runslow, mocker)


RISK_FACTORS = [
    (
        risk_factors.high_systolic_blood_pressure,
        [
            "exposure",
            "exposure_standard_deviation",
            "exposure_distribution_weights",
            "relative_risk",  # Very slow
            "population_attributable_fraction",  # Very slow
        ],
    ),
    (
        risk_factors.low_birth_weight_and_short_gestation,
        [
            "exposure",
            "relative_risk",  # Very slow
            "population_attributable_fraction",  # Very slow
        ],
    ),
]
RISK_FACTOR_MEASURES = [
    "exposure",
    "exposure_standard_deviation",
    "exposure_distribution_weights",
    "relative_risk",
    "population_attributable_fraction",
]


@pytest.mark.parametrize("entity_details", RISK_FACTORS, ids=lambda x: x[0].name)
@pytest.mark.parametrize("measure", RISK_FACTOR_MEASURES, ids=lambda x: x)
@pytest.mark.parametrize(
    "data_type", ["means", "draws", ["means", "draws"]], ids=lambda x: str(x)
)
@pytest.mark.parametrize(
    "mock_gbd", [True, False], ids=lambda x: "mocked" if x else "unmocked"
)
def test_get_measure_risklike(
    entity_details: tuple(RiskFactor, list[str]),
    measure: str,
    data_type: str | list[str],
    mock_gbd: bool,
    runslow: bool,
    mocker: MockerFixture,
):
    """Test get_measure().

    We are parametrizing over various things.
    - entity_details: A tuple of an entity and a list of measures that are applicable to that entity.
    - measure: The possible measures for that entity type (e.g. cause, sequela, etc)
    - data_type: The data type request (means or draws)
    - mock_gbd: Whether to mock calls to vivarium_gbd_access or not. We do this because
        getting the real data tends to take a long time and so we want to only run that
        when requesting a --runslow test. Note that we do not attempt to get more granular
        with slow runs; mocked tests will always run and unmocked tests will only run
        if --runslow is passed

    Notes
    -----
    The --runslow flag is automatically passed into these tests as an argument; we do
    not mark the tests themselves as slow (because we want more granularity, i.e.
    mocked tests are not slow but unmocked tests are).
    """

    # Test-specific fixme skips
    if (
        measure in ["relative_risk", "population_attributable_fraction"]
        and entity_details[0].name == "high_systolic_blood_pressure"
        and not mock_gbd
        and data_type == "draws"
    ):
        pytest.skip("FIXME: [mic-5542] continuous rrs cannot validate")

    run_test(entity_details, measure, data_type, mock_gbd, runslow, mocker)


COVARIATES = [(covariates.systolic_blood_pressure_mmhg, ["estimate"])]
COVARIATE_MEASURES = ["estimate"]


@pytest.mark.parametrize("entity_details", COVARIATES, ids=lambda x: x[0].name)
@pytest.mark.parametrize("measure", COVARIATE_MEASURES, ids=lambda x: x)
@pytest.mark.parametrize(
    "data_type", ["means", "draws", ["means", "draws"]], ids=lambda x: str(x)
)
@pytest.mark.parametrize(
    "mock_gbd", [True, False], ids=lambda x: "mocked" if x else "unmocked"
)
def test_get_measure_covariatelike(
    entity_details: tuple(Covariate, list[str]),
    measure: str,
    data_type: str | list[str],
    mock_gbd: bool,
    runslow: bool,
    mocker: MockerFixture,
):
    """Test get_measure().

    We are parametrizing over various things.
    - entity_details: A tuple of an entity and a list of measures that are applicable to that entity.
    - measure: The possible measures for that entity type (e.g. cause, sequela, etc)
    - data_type: The data type request (means or draws)
    - mock_gbd: Whether to mock calls to vivarium_gbd_access or not. We do this because
        getting the real data tends to take a long time and so we want to only run that
        when requesting a --runslow test. Note that we do not attempt to get more granular
        with slow runs; mocked tests will always run and unmocked tests will only run
        if --runslow is passed

    Notes
    -----
    The --runslow flag is automatically passed into these tests as an argument; we do
    not mark the tests themselves as slow (because we want more granularity, i.e.
    mocked tests are not slow but unmocked tests are).
    """

    run_test(entity_details, measure, data_type, mock_gbd, runslow, mocker)


# TODO [MIC-5550]: Add tests for etiologies and alternative risk factors


####################
# HELPER FUNCTIONS #
####################


# FIXME [MIC-5454]: Move to vivarium_testing_utilities
def is_slow_test_day(slow_test_day: str = SLOW_TEST_DAY) -> bool:
    """Determine if today is the day to run slow/weekly tests.

    Parameters
    ----------
    slow_test_day
        The day to run the weekly tests on. Acceptable values are "Monday",
        "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", or "Sunday".
        Default is "Sunday".

    Notes
    -----
    There is some risk that a test will be inadvertently skipped if there is a
    significant delay between when a pipeline is kicked off and when the test
    itself is run.
    """
    return [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
        "Sunday",
    ][datetime.datetime.today().weekday()] == slow_test_day


def run_test(
    entity_details: tuple[Cause | Sequela | RiskFactor, list[str]],
    measure: str,
    data_type: str | list[str],
    mock_gbd: bool,
    runslow: bool,
    mocker: MockerFixture,
    raise_type: Exception = None,
):
    entity, entity_expected_measures = entity_details

    if not runslow and not mock_gbd:
        pytest.skip("Need --runslow option to run unmocked tests")
    if NO_GBD_ACCESS and not mock_gbd:
        pytest.skip("Need GBD access to run unmocked tests")
    # Only run PAF tests on Sundays since it's so slow.
    if (
        measure in ["population_attributable_fraction", "relative_risk"]
        and measure in entity_expected_measures
        and not mock_gbd
        and not is_slow_test_day()
    ):
        pytest.skip(f"Only run full PAF and RR tests on {SLOW_TEST_DAY}s")

    tester = success_expected if measure in entity_expected_measures else fail_expected
    is_unimplemented = is_not_implemented(data_type, measure)

    if is_unimplemented:  # This should trigger first
        tester = partial(fail_expected, raise_type=DataTypeNotImplementedError)
    elif raise_type:
        tester = partial(fail_expected, raise_type=raise_type)
    else:
        pass  # success is really expected

    if mock_gbd:
        # TODO: Reduce duplicate testing. Since this data is mocked, it is doing the
        # same test for a given measure regardless of the entity.
        if is_unimplemented:
            pytest.skip("Cannot mock data for unimplemented features.")
        if tester == fail_expected:
            pytest.skip("Do mock data for expected failed calls.")
        mocked_funcs = mock_vivarium_gbd_access(
            entity, measure, LOCATION, MOST_RECENT_YEAR, data_type, mocker
        )

    tester(entity, measure, utility_data.get_location_id(LOCATION), data_type)
    if mock_gbd:
        for mocked_func in mocked_funcs:
            assert mocked_func.called_once()


def success_expected(entity, measure, location, data_type):
    df = get_measure(entity, measure, location, years=None, data_type=data_type)
    check_data(entity, measure, df, data_type)


def fail_expected(entity, measure, location, data_type, raise_type=Exception):
    with pytest.raises(raise_type):
        _df = get_measure(entity, measure, location, years=None, data_type=data_type)


def check_data(
    entity: Cause | Sequela | RiskFactor | Covariate,
    measure: str,
    data: pd.DataFrame,
    data_type: str,
) -> None:
    """Check the data returned."""
    if measure in NON_STANDARD_MEASURES:
        assert list(data.columns) == ["value"]
    else:
        assert all(col in data.columns for col in SUPPORTED_DATA_TYPES[data_type])
    assert all(data.notna())
    assert all(data >= 0)
    # Check metadata index values (note that there may be other metadata returned)
    age_bins = get_mocked_age_bins()
    expected_metadata = {
        "location": {LOCATION},
        "sex": {"Male", "Female"},
        "age_start": set(age_bins["age_group_years_start"]),
        "age_end": set(age_bins["age_group_years_end"]),
        "year_start": {MOST_RECENT_YEAR},
        "year_end": {MOST_RECENT_YEAR + 1},
    }
    if not getattr(entity, "by_age", True):
        # Some entities do not have ages
        del expected_metadata["age_start"]
        del expected_metadata["age_end"]
    for idx, expected in expected_metadata.items():
        assert set(data.index.get_level_values(idx)) == expected
